"""Command line interface for pysradb"""

import argparse
import os
import re
import sys
import warnings
from io import StringIO
from textwrap import dedent

import pandas as pd

from . import __version__
from .exceptions import IncorrectFieldException, MissingQueryException
from .geodb import download_geo_matrix, parse_geo_matrix_to_tsv
from .geoweb import GEOweb
from .search import EnaSearch, GeoSearch, SraSearch
from .sradb import SRAdb, download_sradb_file
from .sraweb import SRAweb
from .utils import confirm

pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)

warnings.simplefilter(action="ignore", category=FutureWarning)


class CustomFormatterArgP(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    pass


class ArgParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n" % message)
        self.print_help()
        sys.exit(2)


def pretty_print_df(df, include_header=True):
    if include_header:
        print("\t".join(map(str, list(df.columns))))
    for index, row in df.iterrows():
        print("\t".join(map(str, row.tolist())))


def _print_save_df(df, saveto=None):
    if saveto:
        if saveto.lower().endswith(".csv"):
            #  if the save file format is csv
            df.to_csv(saveto, index=False, header=True)
        else:
            df.to_csv(saveto, index=False, header=True, sep="\t")
    else:
        if df is None:
            print
        elif len(df.index):
            pretty_print_df(df)


###################### metadata ##############################
def metadata(srp_id, assay, desc, detailed, expand, saveto):
    sradb = SRAweb()
    df = sradb.sra_metadata(
        srp_id,
        assay=assay,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


################################################################


################# download ##########################
def download(
    out_dir,
    srx,
    srp,
    geo,
    skip_confirmation,
    col="public_url",
    use_ascp=False,
    threads=1,
):
    if out_dir is None:
        out_dir = os.path.join(os.getcwd(), "pysradb_downloads")
    sradb = SRAweb()
    geoweb = GEOweb()
    # This block is triggered only if no -p or -g arguments are provided.
    # In this case, the input is taken from the pipe and assumed to be SRA, not GEO
    # TODO: at some point, we need to fix this
    if not srp and not geo:
        df = pd.read_csv(sys.stdin, sep="\t")
        sradb.download(
            df=df,
            out_dir=out_dir,
            filter_by_srx=srx,
            skip_confirmation=True,
            use_ascp=use_ascp,
            url_col=col,
            threads=threads,
        )
    # This block is triggered for downloads using the -p argument
    if srp:
        for srp_x in srp:
            metadata = sradb.sra_metadata(srp_x, detailed=True)
            sradb.download(
                df=metadata,
                out_dir=out_dir,
                filter_by_srx=srx,
                skip_confirmation=skip_confirmation,
                use_ascp=use_ascp,
                threads=threads,
            )
    # This block is triggered for downloads using the -g argument
    if geo:
        for geo_x in geo:
            links, root_url = geoweb.get_download_links(geo_x)
            geoweb.download(links=links, root_url=root_url, gse=geo_x, out_dir=out_dir)
    sradb.close()


#########################################################


######################### search #################################
def search(saveto, db, verbosity, return_max, fields):
    if fields["run_description"]:
        verbosity = 1
    if fields["detailed"]:
        verbosity = 3
    try:
        if db == "ena":
            instance = EnaSearch(
                verbosity,
                return_max,
                fields["query"],
                fields["accession"],
                fields["organism"],
                fields["layout"],
                fields["mbases"],
                fields["publication_date"],
                fields["platform"],
                fields["selection"],
                fields["source"],
                fields["strategy"],
                fields["title"],
            )
            instance.search()
        elif db == "geo":
            instance = GeoSearch(
                verbosity,
                return_max,
                fields["query"],
                fields["accession"],
                fields["organism"],
                fields["layout"],
                fields["mbases"],
                fields["publication_date"],
                fields["platform"],
                fields["selection"],
                fields["source"],
                fields["strategy"],
                fields["title"],
                fields["geo_query"],
                fields["geo_dataset_type"],
                fields["geo_entry_type"],
            )
            instance.search()
        else:
            instance = SraSearch(
                verbosity,
                return_max,
                fields["query"],
                fields["accession"],
                fields["organism"],
                fields["layout"],
                fields["mbases"],
                fields["publication_date"],
                fields["platform"],
                fields["selection"],
                fields["source"],
                fields["strategy"],
                fields["title"],
            )
            instance.search()
    except (MissingQueryException, IncorrectFieldException) as e:
        print(e)
        return
    if fields["stats"]:
        instance.show_result_statistics()
    if fields["graphs"]:
        graph_types = tuple(fields["graphs"].split())
        instance.visualise_results(graph_types, False)
    _print_save_df(instance.get_df(), saveto)


def get_geo_search_info():
    print(GeoSearch.info())


####################################################################


######################### gse-to-gsm ###############################
def gse_to_gsm(gse_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gse_to_gsm(
        gse_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


####################################################################


######################## gse-to-srp ################################
def gse_to_srp(gse_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gse_to_srp(
        gse_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


######################################################################


######################### gsm-to-gse #################################
def gsm_to_gse(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_gse(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################ gsm-to-srp ################################
def gsm_to_srp(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srp(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################ gsm-to-srr ################################
def gsm_to_srr(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srr(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################ gsm-to-srs ################################
def gsm_to_srs(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srs(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################# gsm-to-srx ###############################
def gsm_to_srx(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srx(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-gse ##################################
def srp_to_gse(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_gse(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-srr ##################################
def srp_to_srr(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_srr(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-srs ##################################
def srp_to_srs(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_srs(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-srx ##################################
def srp_to_srx(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_srx(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srr-to-gsm ##################################
def srr_to_gsm(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_gsm(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


########################### srr-to-srp ##################################
def srr_to_srp(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_srp(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srr-to-srs ##################################
def srr_to_srs(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_srs(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srr-to-srx ##################################
def srr_to_srx(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_srx(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srs-to-gsm ##################################
def srs_to_gsm(srs_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srs_to_gsm(
        srs_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srs-to-srx ##################################
def srs_to_srx(srs_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srs_to_srx(
        srs_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srx-to-srp ##################################
def srx_to_srp(srx_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srx_to_srp(
        srx_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srx-to-srr ##################################
def srx_to_srr(srx_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srx_to_srr(
        srx_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srx-to-srs ##################################
def srx_to_srs(srx_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srx_to_srs(
        srx_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


def srp_to_pmid(srp_ids, saveto):
    sradb = SRAweb()
    df = sradb.srp_to_pmid(srp_ids)
    _print_save_df(df, saveto)
    sradb.close()


def sra_to_pmid(sra_ids, saveto):
    """Backward compatibility wrapper for sra_to_pmid"""
    sradb = SRAweb()
    df = sradb.sra_to_pmid(sra_ids)
    _print_save_df(df, saveto)
    sradb.close()


def gse_to_pmid(gse_ids, saveto):
    sradb = SRAweb()
    df = sradb.gse_to_pmid(gse_ids)
    _print_save_df(df, saveto)
    sradb.close()


def pmid_to_gse(pmid_ids, saveto):
    sradb = SRAweb()
    df = sradb.pmid_to_gse(pmid_ids)
    _print_save_df(df, saveto)
    sradb.close()


def pmid_to_srp(pmid_ids, saveto):
    sradb = SRAweb()
    df = sradb.pmid_to_srp(pmid_ids)
    _print_save_df(df, saveto)
    sradb.close()


def pmc_to_identifiers(pmc_ids, saveto):
    sradb = SRAweb()
    df = sradb.pmc_to_identifiers(pmc_ids)
    _print_save_df(df, saveto)
    sradb.close()


def pmid_to_identifiers(pmid_ids, saveto):
    sradb = SRAweb()
    df = sradb.pmid_to_identifiers(pmid_ids)
    _print_save_df(df, saveto)
    sradb.close()


def doi_to_gse(doi_ids, saveto):
    sradb = SRAweb()
    df = sradb.doi_to_gse(doi_ids)
    _print_save_df(df, saveto)
    sradb.close()


def doi_to_srp(doi_ids, saveto):
    sradb = SRAweb()
    df = sradb.doi_to_srp(doi_ids)
    _print_save_df(df, saveto)
    sradb.close()


def doi_to_identifiers(doi_ids, saveto):
    sradb = SRAweb()
    df = sradb.doi_to_identifiers(doi_ids)
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### geo-matrix ##################################
def geo_matrix(accession, to_tsv, output_dir):
    # Download the GEO Matrix file
    matrix_file = download_geo_matrix(accession, output_dir=output_dir)
    print(f"Downloaded GEO Matrix file to: {matrix_file}")

    # If --to-tsv is specified, parse the file to TSV
    if to_tsv:
        output_tsv = os.path.join(output_dir, f"{accession}_matrix.tsv")
        df = parse_geo_matrix_to_tsv(matrix_file, output_tsv)
        print(f"Parsed GEO Matrix file to TSV: {output_tsv}")


#########################################################################


def parse_args(args=None):
    """Argument parser"""
    parser = ArgParser(
        description=dedent(
            """\
    pysradb: Query NGS metadata and data from NCBI Sequence Read Archive.
    version: {}.
    Citation: 10.12688/f1000research.18676.1""".format(
                __version__
            )
        ),
        formatter_class=CustomFormatterArgP,
    )

    # pysradb subcommands
    subparsers = parser.add_subparsers(title="subcommands", dest="command")

    # pysradb --version
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )

    # pysradb --citation
    parser.add_argument(
        "--citation",
        action="version",
        version=dedent(
            """
        Choudhary, Saket. "pysradb: A Python Package to Query next-Generation Sequencing Metadata and Data from NCBI Sequence Read Archive." F1000Research, vol. 8, F1000 (Faculty of 1000 Ltd), Apr. 2019, p. 532 (https://f1000research.com/articles/8-532/v1)

        @article{Choudhary2019,
        doi = {10.12688/f1000research.18676.1},
        url = {https://doi.org/10.12688/f1000research.18676.1},
        year = {2019},
        month = apr,
        publisher = {F1000 (Faculty of 1000 Ltd)},
        volume = {8},
        pages = {532},
        author = {Saket Choudhary},
        title = {pysradb: A {P}ython package to query next-generation sequencing metadata and data from {NCBI} {S}equence {R}ead {A}rchive},
        journal = {F1000Research}
        }
        """
        ),
        help="how to cite",
    )

    # pysradb metadata
    subparser = subparsers.add_parser(
        "metadata", help="Fetch metadata for SRA project (SRPnnnn)"
    )
    subparser.add_argument("--saveto", help="Save metadata dataframe to file")
    subparser.add_argument(
        "--assay", action="store_true", help="Include assay type in output"
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed", action="store_true", help="Display detailed metadata table"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id", nargs="+")
    subparser.set_defaults(func=metadata)

    # pysradb download
    subparser = subparsers.add_parser("download", help="Download SRA project (SRPnnnn)")
    subparser.add_argument("--out-dir", help="Output directory root")
    subparser.add_argument("--srx", "-x", help="Download only these SRX(s)", nargs="+")
    subparser.add_argument("--srp", "-p", help="SRP ID", nargs="+")
    subparser.add_argument("--geo", "-g", help="GEO ID", nargs="+")
    subparser.add_argument(
        "--skip-confirmation", "-y", action="store_true", help="Skip confirmation"
    )
    subparser.add_argument(
        "--use_ascp", "-a", action="store_true", help="Use aspera instead of wget"
    )
    subparser.add_argument(
        "--col", help="Specify column to download", default="public_url"
    )
    subparser.add_argument(
        "--threads", "-t", help="Number of threads", default=1, type=int
    )
    subparser.set_defaults(func=download)

    # pysradb search
    subparser = subparsers.add_parser("search", help="Search SRA/ENA for matching text")
    subparser.add_argument(
        "-o", "--saveto", help="Save search result dataframe to file"
    )
    subparser.add_argument(
        "-s",
        "--stats",
        action="store_true",
        help="Displays some useful statistics for the search results.",
    )
    subparser.add_argument(
        "-g",
        "--graphs",
        nargs="?",
        const="all",
        help=(
            "Generates graphs to illustrate the search result. "
            "By default all graphs are generated. \n"
            "Alternatively, select a subset from the options below in a space-separated string:\n"
            "daterange, organism, source, selection, platform, basecount"
        ),
    )
    subparser.add_argument(
        "-d",
        "--db",
        choices=["ena", "geo", "sra"],
        default="sra",
        help="Select the db API (sra, ena, or geo) to query, default = sra.     "
        "Note: pysradb search works slightly differently when db = geo. \n"
        "Please refer to 'pysradb search --geo-info' for more details.",
    )
    subparser.add_argument(
        "-v",
        "--verbosity",
        choices=[0, 1, 2, 3],
        default=2,
        help=(
            "Level of search result details (0, 1, 2 or 3), default = 2\n"
            "0: run accession only\n"
            "1: run accession and experiment title\n"
            "2: accession numbers, titles and sequencing information\n"
            "3: records in 2 and other information such as download url, sample attributes, etc"
        ),
        type=int,
    )
    subparser.add_argument(
        "--run-description",
        action="store_true",
        help="Displays run accessions and descriptions only. Equivalent to --verbosity 1",
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Displays detailed search results. Equivalent to --verbosity 3.",
    )
    subparser.add_argument(
        "-m",
        "--max",
        default=20,
        help="Maximum number of entries to return, default = 20",
        type=int,
    )
    subparser.add_argument(
        "-q",
        "--query",
        nargs="+",
        help="Main query string. Note that if no query is supplied, at least one of the "
        "following flags must be present:",
    )
    subparser.add_argument("-A", "--accession", help="Accession number")
    subparser.add_argument(
        "-O", "--organism", nargs="+", help="Scientific name of the sample organism"
    )
    subparser.add_argument(
        "-L",
        "--layout",
        choices=["SINGLE", "PAIRED"],
        help="Library layout. Accepts either SINGLE or PAIRED",
        type=str.upper,
    )
    subparser.add_argument(
        "-M",
        "--mbases",
        help="Size of the sample rounded to the nearest megabase",
        type=int,
    )
    subparser.add_argument(
        "-D",
        "--publication-date",
        help="Publication date of the run in the format dd-mm-yyyy. If a date range is desired, "
        "enter the start date, followed by end date, separated by a colon ':'.\n "
        "Example: 01-01-2010:31-12-2010",
    )
    subparser.add_argument("-P", "--platform", nargs="+", help="Sequencing platform")
    subparser.add_argument("-E", "--selection", nargs="+", help="Library selection")
    subparser.add_argument("-C", "--source", nargs="+", help="Library source")
    subparser.add_argument(
        "-S", "--strategy", nargs="+", help="Library preparation strategy"
    )
    subparser.add_argument("-T", "--title", nargs="+", help="Experiment title")

    # The following arguments are for GEO DataSets only
    subparser.add_argument(
        "-I",
        "--geo-info",
        action="store_true",
        help="Displays information on how to query GEO DataSets via 'pysradb search --db geo ...', "
        "including accepted inputs for -G/--geo-query, -Y/--geo-dataset-type and -Z/--geo-entry-type. ",
    )
    subparser.add_argument(
        "-G",
        "--geo-query",
        nargs="+",
        help="Main query string for GEO DataSet. This flag is only used when db is set to be geo."
        "Please refer to 'pysradb search --geo-info' for more details.",
    )
    subparser.add_argument(
        "-Y",
        "--geo-dataset-type",
        nargs="+",
        help="GEO DataSet Type. This flag is only used when --db is set to be geo."
        "Please refer to 'pysradb search --geo-info' for more details.",
    )
    subparser.add_argument(
        "-Z",
        "--geo-entry-type",
        nargs="+",
        help="GEO Entry Type. This flag is only used when --db is set to be geo."
        "Please refer to 'pysradb search --geo-info' for more details.",
    )

    subparser.set_defaults(func=search)

    # pysradb gse-to-gsm
    subparser = subparsers.add_parser("gse-to-gsm", help="Get GSM for a GSE")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                         run_accession (SRR),
                                         sample_alias (GSM),
                                         run_alias (GSM_r)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("gse_ids", nargs="+")
    subparser.set_defaults(func=gse_to_gsm)

    # pysradb gse-to-srp
    subparser = subparsers.add_parser("gse-to-srp", help="Get SRP for a GSE")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                           run_accession (SRR),
                                           sample_accession (SRS),
                                           experiment_alias (GSM_),
                                           run_alias (GSM_r),
                                           sample_alias (GSM)]
                                           """,
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("gse_ids", nargs="+")
    subparser.set_defaults(func=gse_to_gsm)

    # pysradb gsm-to-gse
    subparser = subparsers.add_parser("gsm-to-gse", help="Get GSE for a GSM")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                            run_accession (SRR),
                                            sample_alias (GSM),
                                            run_alias (GSM_r)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_gse)

    # pysradb gsm-to-srp
    subparser = subparsers.add_parser("gsm-to-srp", help="Get SRP for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            run_accession (SRR),
                                            experiment_alias (GSM),
                                            sample_alias (GSM),
                                            run_alias (GSM_r),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srp)

    # pysradb gsm-to-srr
    subparser = subparsers.add_parser("gsm-to-srr", help="Get SRR for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            study_accession (SRP),
                                            run_alias (GSM_r),
                                            sample_alias (GSM),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srr)

    # pysradb gsm-to-srs
    subparser = subparsers.add_parser("gsm-to-srs", help="Get SRS for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            run_accession (SRR),
                                            study_accession (SRP),
                                            run_alias (GSM_r),
                                            experiment_alias (GSM),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srs)

    # pysradb gsm-to-srx
    subparser = subparsers.add_parser("gsm-to-srx", help="Get SRX for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            run_accession (SRR),
                                            experiment_alias (GSM),
                                            sample_alias (GSM),
                                            run_alias (GSM_r),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srx)

    # pysradb srp-to-gse
    subparser = subparsers.add_parser("srp-to-gse", help="Get GSE for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [sample_accession, run_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_gse)

    # pysradb srp-to-srr
    subparser = subparsers.add_parser("srp-to-srr", help="Get SRR for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            study_alias (GSE),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            run_alias (GSM_r)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_srr)

    # pysradb srp-to-srs
    subparser = subparsers.add_parser("srp-to-srs", help="Get SRS for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [run_accession (SRR),
                                            study_accession (SRP),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            run_alias (GSM_r),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_srs)

    # pysradb srp-to-srx
    subparser = subparsers.add_parser("srp-to-srx", help="Get SRX for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                            run_accession (SRR),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            run_alias (GSM_r)',
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_srx)

    # pysradb srr-to-gsm
    subparser = subparsers.add_parser("srr-to-gsm", help="Get GSM for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""'Output additional columns: [experiment_accession (SRX),
                                             study_accession (SRP),
                                             run_alias (GSM_r),
                                             sample_alias (GSM_),
                                             experiment_alias (GSM),
                                             study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_gsm)

    # pysradb srr-to-srp
    subparser = subparsers.add_parser("srr-to-srp", help="Get SRP for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""'Output additional columns: [experiment_accession (SRX),
                                             sample_accession (SRS),
                                             run_alias (GSM_r),
                                             experiment_alias (GSM),
                                             sample_alias (GSM_),
                                             study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_srp)

    # pysradb srr-to-srs
    subparser = subparsers.add_parser("srr-to-srs", help="Get SRS for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""'Output additional columns: [experiment_accession (SRX),
                                             study_accession (SRP),
                                             run_alias (GSM_r),
                                             sample_alias (GSM_),
                                             experiment_alias (GSM),
                                             study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_srs)

    # pysradb srr-to-srx
    subparser = subparsers.add_parser("srr-to-srx", help="Get SRX for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                            study_accession (SRP),
                                            run_alias (GSM_r),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_srx)

    # pysradb srs-to-gsm
    subparser = subparsers.add_parser("srs-to-gsm", help="Get GSM for a SRS")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [run_accession, study_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srs_ids", nargs="+")
    subparser.set_defaults(func=srs_to_gsm)

    # pysradb srs-to-srx
    subparser = subparsers.add_parser("srs-to-srx", help="Get SRX for a SRS")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [run_accession, study_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srs_ids", nargs="+")
    subparser.set_defaults(func=srs_to_srx)

    # pysradb srx-to-srp
    subparser = subparsers.add_parser("srx-to-srp", help="Get SRP for a SRX")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [run_accession (SRR),
                                            sample_accession (SRS),
                                            experiment_alias (GSM),
                                            run_alias (GSM_r),
                                            sample_alias (GSM),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srx_ids", nargs="+")
    subparser.set_defaults(func=srx_to_srp)

    # pysradb srx-to-srr
    subparser = subparsers.add_parser("srx-to-srr", help="Get SRR for a SRX")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [sample_accession, study_accession]",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srx_ids", nargs="+")
    subparser.set_defaults(func=srx_to_srr)

    # pysradb srx-to-srs
    subparser = subparsers.add_parser("srx-to-srs", help="Get SRS for a SRX")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [run_accession, study_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srx_ids", nargs="+")
    subparser.set_defaults(func=srx_to_srs)

    # pysradb geo-matrix
    subparser = subparsers.add_parser(
        "geo-matrix", help="Download and parse GEO Matrix files"
    )
    subparser.add_argument(
        "--accession", required=True, help="GEO accession (e.g., GSE234190)"
    )
    subparser.add_argument(
        "--to-tsv", action="store_true", help="Convert the matrix file to TSV format"
    )
    subparser.add_argument(
        "--output-dir",
        default=".",
        help="Output directory (default: current directory)",
    )
    subparser.set_defaults(func=geo_matrix)

    # pysradb srp-to-pmid
    subparser = subparsers.add_parser(
        "srp-to-pmid", help="Get PMIDs for SRP accessions"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srp_ids", nargs="+", help="SRP accession(s)")
    subparser.set_defaults(func=srp_to_pmid)

    # pysradb gse-to-pmid
    subparser = subparsers.add_parser(
        "gse-to-pmid", help="Get PMIDs for GSE accessions"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gse_ids", nargs="+", help="GSE accession(s)")
    subparser.set_defaults(func=gse_to_pmid)

    # pysradb pmid-to-gse
    subparser = subparsers.add_parser(
        "pmid-to-gse", help="Get GSE accessions from PMIDs"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("pmid_ids", nargs="+", help="PMID(s)")
    subparser.set_defaults(func=pmid_to_gse)

    # pysradb pmid-to-srp
    subparser = subparsers.add_parser(
        "pmid-to-srp", help="Get SRP accessions from PMIDs"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("pmid_ids", nargs="+", help="PMID(s)")
    subparser.set_defaults(func=pmid_to_srp)

    # pysradb pmc-to-identifiers
    subparser = subparsers.add_parser(
        "pmc-to-identifiers", help="Extract database identifiers from PMC articles"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("pmc_ids", nargs="+", help="PMC ID(s)")
    subparser.set_defaults(func=pmc_to_identifiers)

    # pysradb pmid-to-identifiers
    subparser = subparsers.add_parser(
        "pmid-to-identifiers", help="Extract database identifiers from PubMed articles"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("pmid_ids", nargs="+", help="PMID(s)")
    subparser.set_defaults(func=pmid_to_identifiers)

    # pysradb doi-to-gse
    subparser = subparsers.add_parser("doi-to-gse", help="Get GSE accessions from DOIs")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("doi_ids", nargs="+", help="DOI(s)")
    subparser.set_defaults(func=doi_to_gse)

    # pysradb doi-to-srp
    subparser = subparsers.add_parser("doi-to-srp", help="Get SRP accessions from DOIs")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("doi_ids", nargs="+", help="DOI(s)")
    subparser.set_defaults(func=doi_to_srp)

    # pysradb doi-to-identifiers
    subparser = subparsers.add_parser(
        "doi-to-identifiers", help="Extract database identifiers from articles via DOI"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("doi_ids", nargs="+", help="DOI(s)")
    subparser.set_defaults(func=doi_to_identifiers)

    args = parser.parse_args(args=None if sys.argv[1:] else ["--help"])
    if args.command == "metadata":
        metadata(
            args.srp_id,
            args.assay,
            args.desc,
            args.detailed,
            args.expand,
            args.saveto,
        )
    elif args.command == "download":
        download(
            args.out_dir,
            args.srx,
            args.srp,
            args.geo,
            args.skip_confirmation,
            args.col,
            args.use_ascp,
            args.threads,
        )
    elif args.command == "search":
        flags = vars(args)
        if flags.pop("geo_info"):
            get_geo_search_info()
        else:
            search(
                flags.pop("saveto"),
                flags.pop("db"),
                flags.pop("verbosity"),
                flags.pop("max"),
                flags,
            )
    elif args.command == "gse-to-gsm":
        gse_to_gsm(args.gse_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gse-to-srp":
        gse_to_srp(args.gse_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-gse":
        gsm_to_gse(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srp":
        gsm_to_srp(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srr":
        gsm_to_srr(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srs":
        gsm_to_srs(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srx":
        gsm_to_srx(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-gse":
        srp_to_gse(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-srr":
        srp_to_srr(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-srs":
        srp_to_srs(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-srx":
        srp_to_srx(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-gsm":
        srr_to_gsm(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-srp":
        srr_to_srp(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-srs":
        srr_to_srs(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-srx":
        srr_to_srx(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srs-to-gsm":
        srs_to_gsm(args.srs_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srs-to-srx":
        srs_to_srx(args.srs_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srx-to-srp":
        srx_to_srp(args.srx_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srx-to-srr":
        srx_to_srr(args.srx_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srx-to-srs":
        srx_to_srs(args.srx_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "geo-matrix":
        geo_matrix(args.accession, args.to_tsv, args.output_dir)
    elif args.command == "srp-to-pmid":
        srp_to_pmid(args.srp_ids, args.saveto)
    elif args.command == "gse-to-pmid":
        gse_to_pmid(args.gse_ids, args.saveto)
    elif args.command == "pmid-to-gse":
        pmid_to_gse(args.pmid_ids, args.saveto)
    elif args.command == "pmid-to-srp":
        pmid_to_srp(args.pmid_ids, args.saveto)
    elif args.command == "pmc-to-identifiers":
        pmc_to_identifiers(args.pmc_ids, args.saveto)
    elif args.command == "pmid-to-identifiers":
        pmid_to_identifiers(args.pmid_ids, args.saveto)
    elif args.command == "doi-to-gse":
        doi_to_gse(args.doi_ids, args.saveto)
    elif args.command == "doi-to-srp":
        doi_to_srp(args.doi_ids, args.saveto)
    elif args.command == "doi-to-identifiers":
        doi_to_identifiers(args.doi_ids, args.saveto)


if __name__ == "__main__":
    parse_args(sys.argv[1:])



"""Utility function to download data"""

import hashlib
import math
import os
import shutil
import sys
import warnings
from ftplib import FTP
from urllib.parse import urlparse

import numpy as np
import requests
from tqdm.autonotebook import tqdm

from .utils import requests_3_retries

warnings.simplefilter(action="ignore", category=FutureWarning)
import pandas as pd

tqdm.pandas()


def _get_ftp_file_size(url):
    """Get file size from FTP server.

    Parameters
    ----------
    url : str
        FTP URL

    Returns
    -------
    size : int
        File size in bytes, or 0 if unable to determine
    """
    try:
        parsed = urlparse(url)
        ftp = FTP(parsed.netloc)
        ftp.login()
        size = ftp.size(parsed.path)
        ftp.quit()
        return size if size is not None else 0
    except Exception:
        return 0


def _download_ftp_file(
    url, file_path, timeout=10, block_size=1024 * 1024, show_progress=False
):
    """Download file from FTP server.

    Parameters
    ----------
    url : str
        FTP URL
    file_path : str
        Local file path to store the downloaded file
    timeout : int
        Timeout in seconds
    block_size : int
        Block size for downloading
    show_progress : bool
        Show progress bar
    """
    parsed = urlparse(url)
    tmp_file_path = file_path + ".part"

    # Check if partial file exists
    first_byte = os.path.getsize(tmp_file_path) if os.path.exists(tmp_file_path) else 0
    file_mode = "ab" if first_byte else "wb"

    try:
        ftp = FTP(parsed.netloc, timeout=timeout)
        ftp.login()

        file_size = ftp.size(parsed.path)
        if file_size is None:
            file_size = -1

        if show_progress and file_size > 0:
            desc = "Downloading {}".format(url.split("/")[-1])
            pbar = tqdm(
                total=file_size,
                initial=first_byte,
                unit="B",
                unit_scale=True,
                desc=desc,
            )

        with open(tmp_file_path, file_mode) as f:
            if first_byte > 0:
                ftp.voidcmd(f"REST {first_byte}")

            def callback(data):
                f.write(data)
                if show_progress and file_size > 0:
                    pbar.update(len(data))

            ftp.retrbinary(f"RETR {parsed.path}", callback, blocksize=block_size)

        if show_progress and file_size > 0:
            pbar.close()

        ftp.quit()

        if file_size == -1 or file_size == os.path.getsize(tmp_file_path):
            shutil.move(tmp_file_path, file_path)
        else:
            raise Exception(
                f"Download incomplete: expected {file_size} bytes, got {os.path.getsize(tmp_file_path)} bytes"
            )

    except Exception as e:
        if show_progress and "pbar" in locals():
            pbar.close()
        raise Exception(f"FTP download failed: {e}")


def millify(n):
    """Convert integer to human readable format.

    Parameters
    ----------
    n : int

    Returns
    -------
    millidx : str
              Formatted integer
    """
    millnames = ["", " KB", " MB", " GB", " TB"]
    # Source: http://stackoverflow.com/a/3155023/756986
    n = float(n)
    millidx = max(
        0,
        min(
            len(millnames) - 1, int(math.floor(0 if n == 0 else math.log10(abs(n)) / 3))
        ),
    )

    return "{:.1f}{}".format(n / 10 ** (3 * millidx), millnames[millidx])


def get_file_size(row, url_col):
    """Get size of file to be downloaded.

    Parameters
    ----------
    row: pd.DataFrame row

    url_col: str
        url_column

    Returns
    -------
    content_length: int
    """
    if row[url_col] is not None:
        url = row[url_col]
    else:
        url = row.download_url
    if url is pd.NA:
        return 0
    if not isinstance(url, str):
        return 0
    if url.startswith("ftp."):
        url = "ftp://" + url

    if url.startswith("ftp://"):
        return _get_ftp_file_size(url)

    try:
        r = requests_3_retries().head(url)
        size = int(r.headers["content-length"])
        r.raise_for_status()
    except requests.exceptions.Timeout:
        sys.exit(f"Connection to {url} has timed out. Please retry.")
    except requests.exceptions.HTTPError:
        print(
            f"The download URL:  {url}  is likely invalid.\n"
            f"Removing {row.run_accession} from the download list\n",
            flush=True,
        )
        return np.NaN
    except KeyError:
        print("Key error for: " + url, flush=True)
        return 0
    return size


def md5_validate_file(file_path, md5_hash):
    """Check file containt against an MD5.

    Parameters
    ----------
    file_path: string
               Path to file
    md5_hash: string
             Expected md5 hash

    Returns
    -------
    valid: bool
           True if expected and observed md5 match
    """
    observed_md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        while True:
            # read 1MB
            chunk = f.read(1000 * 1000)
            if not chunk:
                break
            observed_md5.update(chunk)
    return observed_md5.hexdigest() == md5_hash


def download_file(
    url,
    file_path,
    md5_hash=None,
    timeout=10,
    block_size=1024 * 1024,
    show_progress=False,
):
    """Resumable download.
    Expect the server to support byte ranges.

    Parameters
    ----------
    url: string
         URL
    file_path: string
               Local file path to store the downloaded file
    md5_hash: string
              Expected MD5 string of downloaded file
    timeout: int
             Seconds to wait before terminating request
    block_size: int
                Chunkx of bytes to read (default: 1024 * 1024 = 1MB)
    show_progress: bool
                   Show progress bar
    """
    if url.startswith("ftp."):
        url = "ftp://" + url

    if os.path.exists(file_path) and os.path.getsize(file_path):
        return

    if url.startswith("ftp://"):
        _download_ftp_file(url, file_path, timeout, block_size, show_progress)
        # if there's a hash value, validate the file
        if md5_hash and not md5_validate_file(file_path, md5_hash):
            raise Exception("Error validating the file against its MD5 hash")
        return

    session = requests
    tmp_file_path = file_path + ".part"
    first_byte = os.path.getsize(tmp_file_path) if os.path.exists(tmp_file_path) else 0
    file_mode = "ab" if first_byte else "wb"
    file_size = -1
    try:
        file_size = int(session.head(url).headers["Content-length"])
        headers = {"Range": "bytes=%s-" % first_byte}
        r = session.get(url, headers=headers, stream=True)
        if show_progress:
            desc = "Downloading {}".format(url.split("/")[-1])
            pbar = tqdm(
                total=file_size,
                initial=first_byte,
                unit="B",
                unit_scale=True,
                desc=desc,
            )
        with open(tmp_file_path, file_mode) as f:
            for chunk in r.iter_content(chunk_size=block_size):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    if show_progress:
                        pbar.update(len(chunk))
        if show_progress:
            pbar.close()
    except IOError as e:
        sys.stderr.write("IO Error - {}\n".format(e))
    finally:
        # Move the temp file to desired location
        if os.path.exists(tmp_file_path):
            actual_size = os.path.getsize(tmp_file_path)
            if file_size == actual_size:
                if md5_hash and not md5_validate_file(tmp_file_path, md5_hash):
                    raise Exception("Error validating the file against its MD5 hash")
                shutil.move(tmp_file_path, file_path)
            elif file_size == -1:
                # Server didn't provide Content-Length, move the file anyway
                shutil.move(tmp_file_path, file_path)
            else:
                print(
                    f"Warning: File size mismatch for {url}. Expected: {file_size}, Got: {actual_size}"
                )
                if actual_size > 0:
                    shutil.move(tmp_file_path, file_path)



"""Utilities to interact with SRA online"""

import concurrent.futures
import os
import re
import sys
import time
import warnings
from collections import OrderedDict
from json.decoder import JSONDecodeError
from xml.parsers.expat import ExpatError

import numpy as np
import pandas as pd
import requests
import xmltodict

from .sradb import SRAdb

warnings.simplefilter(action="ignore", category=FutureWarning)

from xml.sax.saxutils import escape


def xmlescape(data):
    return escape(data, entities={"'": "&apos;", '"': "&quot;"})


def _make_hashable(obj):
    """Convert unhashable types to hashable ones for pandas operations"""
    if isinstance(obj, (OrderedDict, dict)):
        # Extract text content from XML parsed dict/OrderedDict
        if "#text" in obj:
            return obj["#text"]  # Extract the actual text content
        elif len(obj) == 1 and "@xmlns" in obj:
            return pd.NA  # Handle xmlns-only dicts as missing data
        else:
            # Fallback to string representation for other dict structures
            return str(obj)
    elif isinstance(obj, list):
        # Convert list to tuple
        return tuple(_make_hashable(item) for item in obj)
    else:
        return obj


def _order_first(df, column_order_list):
    columns = column_order_list + [
        col for col in df.columns.tolist() if col not in column_order_list
    ]
    # check if all columns do exist in the dataframe
    if len(set(columns).intersection(df.columns)) == len(columns):
        df = df.loc[:, columns]
    df = df.mask(df.map(str).eq("[]"))
    # Filter out XML namespace artifacts
    df = df.replace(regex=r"^@xmlns.*", value=pd.NA).infer_objects(copy=False)
    df = df.fillna(pd.NA)
    return df


def _retry_response(base_url, payload, key, max_retries=10):
    """Rerty fetching esummary if API rate limit exceeeds"""
    for index, _ in enumerate(range(max_retries)):
        try:
            request = requests.get(base_url, params=OrderedDict(payload))
            response = request.json()
            results = response[key]
            return response
        except KeyError:
            # sleep for increasing times
            time.sleep(index + 1)
            continue
    raise RuntimeError("Failed to fetch esummary. API rate limit exceeded.")


def get_retmax(n_records, retmax=500):
    """Get retstart and retmax till n_records are exhausted"""
    for i in range(0, n_records, retmax):
        yield i


class SRAweb(SRAdb):
    def __init__(self, api_key=None):
        """
        Initialize a SRAwebdb.

        Parameters
        ----------

        api_key: string
                 API key for ncbi eutils.
        """
        self.base_url = dict()
        self.base_url["esummary"] = (
            "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi"
        )
        self.base_url["esearch"] = (
            "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
        )
        self.base_url["efetch"] = (
            "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"
        )

        self.ena_fastq_search_url = "https://www.ebi.ac.uk/ena/portal/api/filereport"
        self.ena_params = [("result", "read_run"), ("fields", "fastq_ftp")]

        self.esearch_params = {}
        self.esearch_params["sra"] = [
            ("db", "sra"),
            ("usehistory", "n"),
            ("retmode", "json"),
        ]
        self.esearch_params["geo"] = [
            ("db", "gds"),
            ("usehistory", "n"),
            ("retmode", "json"),
        ]
        self.efetch_params = [
            ("db", "sra"),
            ("usehistory", "n"),
            ("retmode", "runinfo"),
        ]

        if api_key is not None:
            self.esearch_params["sra"].append(("api_key", str(api_key)))
            self.esearch_params["geo"].append(("api_key", str(api_key)))
            self.efetch_params.append(("api_key", str(api_key)))
            self.sleep_time = 1 / 10
        else:
            self.sleep_time = 1 / 3

    @staticmethod
    def format_xml(string):
        """Create a fake root to make 'string' a valid xml

        Parameters
        ----------
        string: str

        Returns
        --------
        xml: str
        """
        # string = unescape(string.strip())
        string = string.strip()
        return "<root>" + string + "</root>"

    @staticmethod
    def xml_to_json(xml):
        """Convert xml to json.

        Parameters
        ----------
        xml: str
             Input XML

        Returns
        -------
        xml_dict: dict
                  Parsed xml as dict
        """
        try:
            xmldict = xmltodict.parse(
                xml, process_namespaces=False, dict_constructor=OrderedDict
            )
            json = xmldict["root"]
        except ExpatError:
            raise RuntimeError("Unable to parse xml: {}".format(xml))
        return json

    def fetch_ena_fastq(self, srp):
        """Fetch FASTQ records from ENA (EXPERIMENTAL)

        Parameters
        ----------
        srp: string
             Srudy accession

        Returns
        -------
        srr_url: list
                 List of SRR fastq urls
        """
        payload = self.ena_params.copy()
        payload += [("accession", srp)]
        request = requests.get(self.ena_fastq_search_url, params=OrderedDict(payload))
        request_text = request.text.strip()
        urls = []
        for line in request_text.split("\n"):
            if "fastq_ftp" in line:
                continue
            line = line.strip()
            line_split = line.split("\t")
            if len(line_split) != 2:
                continue
            url, srr = line.split("\t")
            # sometimes this needs to be flipped
            if "sra.ebi.ac.uk" in srr:
                url, srr = srr, url
            http_url = "http://{}".format(url)
            ftp_url = url.replace("ftp.sra.ebi.ac.uk/", "era-fasp@fasp.sra.ebi.ac.uk:")
            urls += [(srr, http_url, ftp_url)]

        # Paired end case
        def _handle_url_split(url_split):
            url1_1 = pd.NA
            url1_2 = pd.NA
            for url_temp in url_split:
                if "_1.fastq.gz" in url_temp:
                    url1_1 = url_temp
                elif "_2.fastq.gz" in url_temp:
                    url1_2 = url_temp
            return url1_1, url1_2

        if ";" in request_text:
            urls_expanded = []
            for srr, url1, url2 in urls:
                # strip _1, _2
                srr = srr.split("_")[0]
                if ";" in url1:
                    url1_split = url1.split(";")
                    if len(url1_split) == 2:
                        url1_1, url1_2 = url1_split
                    else:
                        # warnings.warn('ignoring extra urls found for paired end accession')
                        url1_1, url1_2 = _handle_url_split(url1_split)
                    url1_2 = "http://{}".format(url1_2)
                    url2_split = url2.split(";")
                    if len(url2_split) == 2:
                        url2_1, url2_2 = url2_split
                    else:
                        # warnings.warn('ignoring extra urls found for paired end accession')
                        url2_1, url2_2 = _handle_url_split(url2_split)
                else:
                    url1_1 = url1
                    url2_1 = url2
                    url1_2 = ""
                    url2_2 = ""
                urls_expanded.append((srr, url1_1, url1_2, url2_1, url2_2))
            df = pd.DataFrame(
                urls_expanded,
                columns=[
                    "run_accession",
                    "ena_fastq_http_1",
                    "ena_fastq_http_2",
                    "ena_fastq_ftp_1",
                    "ena_fastq_ftp_2",
                ],
            ).sort_values(by="run_accession")
            return df
        else:
            return pd.DataFrame(
                urls, columns=["run_accession", "ena_fastq_http", "ena_fastq_ftp"]
            ).sort_values(by="run_accession")

    def create_esummary_params(self, esearchresult, db="sra"):
        query_key = esearchresult["querykey"]
        webenv = esearchresult["webenv"]
        retstart = esearchresult["retstart"]

        # TODO this should be adaptive to build
        # upon using the 'count' result in esearch result,
        # Currently only supports a max of 500 records.
        # retmax = esearchresult["retmax"]
        retmax = 500

        return [
            ("query_key", query_key),
            ("WebEnv", webenv),
            ("retstart", retstart),
            ("retmax", retmax),
        ]

    def get_esummary_response(self, db, term, usehistory="y"):
        assert db in ["sra", "geo"]

        payload = self.esearch_params[db].copy()
        if isinstance(term, list):
            term = " OR ".join(term)
        payload += [("term", term)]
        request = requests.post(self.base_url["esearch"], data=OrderedDict(payload))
        try:
            esearch_response = request.json()
        except JSONDecodeError:
            sys.stderr.write(
                "Unable to parse esummary response json: {}{}. Will retry once.".format(
                    request.text, os.linesep
                )
            )
            retry_after = request.headers.get("Retry-After", 1)
            time.sleep(int(retry_after))
            request = requests.post(self.base_url["esearch"], data=OrderedDict(payload))
            try:
                esearch_response = request.json()
            except JSONDecodeError as e:
                error_msg = (
                    "Unable to parse esummary response json: {}{}. Aborting.".format(
                        request.text, os.linesep
                    )
                )
                sys.stderr.write(error_msg)
                raise ValueError(error_msg) from e

            # retry again

        if "esummaryresult" in esearch_response:
            print("No result found")
            return
        if "error" in esearch_response:
            # API rate limite exceeded
            esearch_response = _retry_response(
                self.base_url["esearch"], payload, "esearchresult"
            )

        n_records = int(esearch_response["esearchresult"]["count"])

        results = {}
        for retstart in get_retmax(n_records):
            payload = self.esearch_params[db].copy()
            payload += self.create_esummary_params(esearch_response["esearchresult"])
            payload = OrderedDict(payload)
            payload["retstart"] = retstart
            request = requests.get(
                self.base_url["esummary"], params=OrderedDict(payload)
            )
            try:
                response = request.json()
            except JSONDecodeError:
                time.sleep(1)
                response = _retry_response(self.base_url["esummary"], payload, "result")

            if "error" in response:
                # API rate limite exceeded
                response = _retry_response(self.base_url["esummary"], payload, "result")
            if retstart == 0:
                results = response["result"]
            else:
                result = response["result"]
                for key, value in result.items():
                    if key in list(results.keys()):
                        results[key] += value
                    else:
                        results[key] = value
        return results

    def get_efetch_response(self, db, term, usehistory="y"):
        assert db in ["sra", "geo"]

        payload = self.esearch_params[db].copy()
        if isinstance(term, list):
            term = " OR ".join(term)
        payload += [("term", term)]

        request = requests.get(self.base_url["esearch"], params=OrderedDict(payload))
        esearch_response = request.json()
        if "esummaryresult" in esearch_response:
            print("No result found")
            return
        if "error" in esearch_response:
            # API rate limite exceeded
            esearch_response = _retry_response(
                self.base_url["esearch"], payload, "esearchresult"
            )

        n_records = int(esearch_response["esearchresult"]["count"])

        results = {}
        for retstart in get_retmax(n_records):
            payload = self.efetch_params.copy()
            payload += self.create_esummary_params(esearch_response["esearchresult"])
            payload = OrderedDict(payload)
            payload["retstart"] = retstart
            request = requests.get(self.base_url["efetch"], params=OrderedDict(payload))
            request_text = request.text.strip()
            try:
                request_json = request.json()
            except:
                request_json = {}  # eval(request_text)

            if "error" in request_json:
                # print("Encountered: {}".format(request_json))
                # print("Headers: {}".format(request.headers))
                # Handle API-rate limit exceeding
                try:
                    retry_after = request.headers["Retry-After"]
                except KeyError:
                    if request_json["error"] == "error forwarding request":
                        error_msg = "Encountered error while making request.\n"
                        sys.stderr.write(error_msg)
                        raise RuntimeError(error_msg.strip())
                time.sleep(int(retry_after))
                # try again
                request = requests.get(
                    self.base_url["efetch"], params=OrderedDict(payload)
                )
                request_text = request.text.strip()
                try:
                    request_json = request.json()
                    if request_json["error"] == "error forwarding request":
                        sys.stderr.write("Encountered error while making request.\n")
                        return
                except:
                    request_json = {}  # eval(request_text)
            try:
                xml_response = xmltodict.parse(
                    request_text, process_namespaces=False, dict_constructor=OrderedDict
                )

                exp_response = xml_response.get("EXPERIMENT_PACKAGE_SET", {})
                response = exp_response.get("EXPERIMENT_PACKAGE", {})
            except ExpatError as e:
                error_msg = "Unable to parse xml: {}{}".format(request_text, os.linesep)
                sys.stderr.write(error_msg)
                raise ValueError(error_msg.strip()) from e
            if not response:
                error_msg = "Unable to parse xml response. Received: {}{}".format(
                    xml_response, os.linesep
                )
                sys.stderr.write(error_msg)
                raise ValueError(error_msg.strip())
            if retstart == 0:
                results = response
            else:
                result = response
                for value in result:
                    results.append(value)
            time.sleep(self.sleep_time)
        return results

    def sra_metadata(
        self,
        srp,
        sample_attribute=False,
        detailed=False,
        expand_sample_attributes=False,
        output_read_lengths=False,
        include_pmids=False,
        **kwargs,
    ):
        esummary_result = self.get_esummary_response("sra", srp)
        try:
            uids = esummary_result["uids"]
        except KeyError:
            return None

        exps_xml = OrderedDict()
        runs_xml = OrderedDict()

        exps_json = OrderedDict()
        runs_json = OrderedDict()
        for uid in uids:
            exps_xml[uid] = self.format_xml(esummary_result[uid]["expxml"])
            runs_xml[uid] = self.format_xml(esummary_result[uid]["runs"])

        for uid in uids:
            exps_json[uid] = self.xml_to_json(exps_xml[uid])
            runs_json[uid] = self.xml_to_json(runs_xml[uid])

        sra_record = []
        for uid, run_json in runs_json.items():
            exp_json = exps_json[uid]
            exp_summary = exp_json["Summary"]
            exp_title = exp_summary.get("Title", pd.NA)
            exp_platform = exp_summary.get("Platform", {})
            statistics = exp_summary.get("Statistics", {})
            if isinstance(exp_platform, OrderedDict):
                exp_platform_model = exp_platform.get("@instrument_model", pd.NA)
                exp_platform_desc = exp_platform.get("#text", pd.NA)
            else:
                exp_platform_model = pd.NA
                exp_platform_desc = pd.NA

            exp_total_runs = statistics.get("@total_runs", pd.NA)
            exp_total_spots = statistics.get("@total_spots", pd.NA)
            exp_total_size = statistics.get("@total_size", pd.NA)

            # experiment_accession
            exp_ID = exp_json["Experiment"]["@acc"]
            # experiment_title
            exp_name = exp_json["Experiment"]["@name"]

            exp_organism = exp_json.get("Organism", pd.NA)
            exp_organism_name = pd.NA
            exp_taxid = pd.NA
            if isinstance(exp_organism, dict):
                exp_organism_name = exp_organism.get("@ScientificName", pd.NA)
                exp_taxid = exp_organism["@taxid"]

            exp_instrument = list(exp_json["Instrument"].values())[0]

            exp_sample = exp_json["Sample"]
            # sample_accession
            exp_sample_ID = exp_sample["@acc"]
            # sample_title
            exp_sample_name = exp_sample["@name"]

            exp_library_descriptor = exp_json["Library_descriptor"]
            # library_strategy
            exp_library_strategy = exp_library_descriptor["LIBRARY_STRATEGY"]
            if isinstance(exp_library_strategy, dict):
                exp_library_strategy = exp_library_strategy["#text"]
            # library_source
            exp_library_source = exp_library_descriptor["LIBRARY_SOURCE"]
            if isinstance(exp_library_source, dict):
                exp_library_source = exp_library_source["#text"]
            # library_selection
            exp_library_selection = exp_library_descriptor["LIBRARY_SELECTION"]
            if isinstance(exp_library_selection, dict):
                exp_library_selection = exp_library_selection["#text"]
            # library_name
            exp_library_name = exp_library_descriptor.get("LIBRARY_NAME", "")
            if isinstance(exp_library_selection, dict):
                exp_library_name = exp_library_name["#text"]
            # library_layout
            exp_library_layout = list(exp_library_descriptor["LIBRARY_LAYOUT"].keys())[
                0
            ]
            # biosample
            exp_biosample = exp_json.get("Biosample", pd.NA)
            # bioproject
            exp_bioproject = exp_json.get("Bioproject", pd.NA)

            experiment_record = OrderedDict()
            experiment_record["study_accession"] = exp_json["Study"]["@acc"]
            experiment_record["study_title"] = exp_json["Study"]["@name"]
            experiment_record["experiment_accession"] = exp_ID
            experiment_record["experiment_title"] = exp_name
            experiment_record["experiment_desc"] = exp_title

            experiment_record["organism_taxid"] = exp_taxid
            experiment_record["organism_name"] = exp_organism_name

            experiment_record["library_name"] = exp_library_name
            experiment_record["library_strategy"] = exp_library_strategy
            experiment_record["library_source"] = exp_library_source
            experiment_record["library_selection"] = exp_library_selection
            experiment_record["library_layout"] = exp_library_layout
            experiment_record["sample_accession"] = exp_sample_ID
            experiment_record["sample_title"] = exp_sample_name
            experiment_record["biosample"] = exp_biosample
            experiment_record["bioproject"] = exp_bioproject
            experiment_record["instrument"] = exp_instrument
            experiment_record["instrument_model"] = exp_platform_model
            experiment_record["instrument_model_desc"] = exp_platform_desc
            experiment_record["total_spots"] = exp_total_spots
            experiment_record["total_size"] = exp_total_size
            if not run_json:
                # Sometimes the run_accession is not populated by NCBI:
                # df2 = self.srx_to_srr(exp_ID)
                # extra_fields = set(experiment_record.keys()).difference(df2.columns.tolist())
                # for idx, row in df2.iterrows():
                #    for field in extra_fields:
                #        experiment_record[field] = row[field]
                sra_record.append(experiment_record)
                continue
            runs = run_json["Run"]
            if not isinstance(runs, list):
                runs = [runs]
            for run_record in runs:
                run_accession = run_record["@acc"]
                run_total_spots = run_record["@total_spots"]
                run_total_bases = run_record["@total_bases"]

                experiment_record["run_accession"] = run_accession
                experiment_record["run_total_spots"] = run_total_spots
                experiment_record["run_total_bases"] = run_total_bases

                sra_record.append(experiment_record.copy())

        # TODO: the detailed call below does redundant operations
        # the code above this can be completeley done away with

        # Convert any unhashable types to hashable ones before creating DataFrame
        hashable_records = []
        for record in sra_record:
            hashable_record = {k: _make_hashable(v) for k, v in record.items()}
            hashable_records.append(hashable_record)

        metadata_df = pd.DataFrame(hashable_records).drop_duplicates()
        if "run_accession" in metadata_df.columns:
            metadata_df = metadata_df.sort_values(by="run_accession")
        metadata_df.columns = [x.lower().strip() for x in metadata_df.columns]
        # Filter out XML namespace artifacts and replace with NA
        metadata_df = metadata_df.replace(
            regex=r"^@xmlns.*", value=pd.NA
        ).infer_objects(copy=False)
        if not detailed:
            return metadata_df

        time.sleep(self.sleep_time)
        efetch_result = self.get_efetch_response("sra", srp)
        if not isinstance(efetch_result, list):
            if efetch_result:
                efetch_result = [efetch_result]
            else:
                return None

        detailed_records = []
        for record in efetch_result:
            if "SAMPLE" in record.keys() and "SAMPLE_ATTRIBUTES" in record["SAMPLE"]:
                sample_attributes = record["SAMPLE"]["SAMPLE_ATTRIBUTES"][
                    "SAMPLE_ATTRIBUTE"
                ]
            else:
                sample_attributes = []
            if isinstance(sample_attributes, OrderedDict):
                sample_attributes = [sample_attributes]
            exp_record = record["EXPERIMENT"]
            exp_attributes = exp_record.get("EXPERIMENT_ATTRIBUTES", {})
            run_sets = record["RUN_SET"].get("RUN", [])

            if not isinstance(run_sets, list):
                run_sets = [run_sets]

            for run_set in run_sets:
                detailed_record = OrderedDict()
                if not run_json:
                    # Add experiment accession if no run info found earlier
                    detailed_record["experiment_accession"] = exp_record["@accession"]
                # detailed_record["experiment_title"] = exp_record["TITLE"]
                for key, values in exp_attributes.items():
                    key = key.lower()
                    for value_x in values:
                        if not isinstance(value_x, dict):
                            continue
                        tag = value_x["TAG"].lower()
                        value = value_x["VALUE"] if "VALUE" in value_x else None
                        detailed_record[tag] = value
                lib_record = exp_record["DESIGN"]["LIBRARY_DESCRIPTOR"]
                for key, value in lib_record.items():
                    key = key.lower()
                    if key == "library_layout":
                        value = list(value.keys())[0]
                    elif key == "library_construction_protocol":
                        continue
                    # detailed_record[key] = value

                detailed_record["run_accession"] = run_set["@accession"]
                detailed_record["run_alias"] = run_set["@alias"]
                sra_files = run_set.get("SRAFiles", {})
                sra_files = sra_files.get("SRAFile", {})
                if isinstance(sra_files, OrderedDict):
                    # detailed_record["sra_url"] = sra_files.get("@url", pd.NA)
                    if "Alternatives" in sra_files.keys():
                        alternatives = sra_files["Alternatives"]
                        if not isinstance(alternatives, list):
                            alternatives = [alternatives]
                        for alternative in alternatives:
                            org = alternative["@org"].lower()
                            for key in alternative.keys():
                                if key == "@org":
                                    continue
                                detailed_record[
                                    "{}_{}".format(org, key.replace("@", ""))
                                ] = alternative[key]

                else:
                    for sra_file in sra_files:
                        # Multiple download URLs
                        # Use the one where the download filename corresponds to the SRR
                        cluster = sra_file.get("@cluster", None).lower().strip()
                        if cluster is None:
                            continue
                        for key in sra_file.keys():
                            if key == "@cluster":
                                continue
                            if key == "Alternatives":
                                # Example: SRP184142
                                alternatives = sra_file["Alternatives"]
                                if not isinstance(alternatives, list):
                                    alternatives = [alternatives]
                                for alternative in alternatives:
                                    org = alternative["@org"].lower()
                                    for key in alternative.keys():
                                        if key == "@org":
                                            continue
                                        detailed_record[
                                            "{}_{}".format(org, key.replace("@", ""))
                                        ] = alternative[key]
                            else:
                                detailed_record[
                                    "{}_{}".format(cluster, key.replace("@", ""))
                                ] = sra_file[key]

                expt_ref = run_set["EXPERIMENT_REF"]
                detailed_record["experiment_alias"] = expt_ref.get("@refname", "")
                # detailed_record["run_total_bases"] = run_set["@total_bases"]
                # detailed_record["run_total_spots"] = run_set["@total_spots"]
                for sample_attribute in sample_attributes:
                    dict_values = list(sample_attribute.values())
                    if len(dict_values) > 1:
                        detailed_record[dict_values[0]] = dict_values[1]
                    else:
                        # TODO: Investigate why these fields have just the key
                        # but no value
                        pass
                detailed_records.append(detailed_record)
        detailed_record_df = pd.DataFrame(detailed_records).drop_duplicates()
        if (
            "run_accession" in metadata_df.keys()
            and "run_accession" in detailed_record_df.keys()
        ):
            metadata_df = metadata_df.merge(
                detailed_record_df, on="run_accession", how="outer"
            )
        elif "experiment_accession" in detailed_record_df.keys():
            metadata_df = metadata_df.merge(
                detailed_record_df, on="experiment_accession", how="outer"
            )

        metadata_df = metadata_df[metadata_df.columns.dropna()]
        metadata_df = metadata_df.drop_duplicates()
        metadata_df = metadata_df.replace(r"^\s*$", pd.NA, regex=True)
        ena_cols = [
            "ena_fastq_http",
            "ena_fastq_http_1",
            "ena_fastq_http_2",
            "ena_fastq_ftp",
            "ena_fastq_ftp_1",
            "ena_fastq_ftp_2",
        ]
        empty_df = pd.DataFrame(columns=ena_cols)
        metadata_df = pd.concat((metadata_df, empty_df), axis=0)

        if "run_accession" in metadata_df.columns:
            metadata_df = metadata_df.set_index("run_accession")
        # multithreading lookup on ENA, since a lot of time is spent waiting
        # for its reply
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            # load our function calls into a list of futures
            futures = [
                executor.submit(self.fetch_ena_fastq, srp)
                for srp in metadata_df.study_accession.unique()
            ]
            # now proceed synchronously
            for future in concurrent.futures.as_completed(futures):
                ena_results = future.result()
                if ena_results.shape[0]:
                    ena_results = ena_results.set_index("run_accession")
                    metadata_df.update(ena_results)
        metadata_df = metadata_df.reset_index()
        metadata_df = metadata_df.fillna(pd.NA)
        metadata_df.columns = [x.lower().strip() for x in metadata_df.columns]

        # Add PMID column when detailed=True and include_pmids=True
        if include_pmids:
            try:
                sra_accessions = [srp] if isinstance(srp, str) else srp
                pmid_df = self.sra_to_pmid(sra_accessions)

                if pmid_df is not None and not pmid_df.empty:
                    pmid_map = {}
                    for _, row in pmid_df.iterrows():
                        study_acc = row.get("sra_accession", None)
                        pmid = row.get("pmid")
                        if not pd.isna(pmid):
                            if study_acc not in pmid_map:
                                pmid_map[study_acc] = []
                            pmid_map[study_acc].append(str(pmid))

                    metadata_df["pmid"] = metadata_df.apply(
                        lambda row: ",".join(
                            pmid_map.get(row.get("study_accession", ""), [""])
                        ),
                        axis=1,
                    )
                    metadata_df["pmid"] = metadata_df["pmid"].replace("", pd.NA)
                else:
                    metadata_df["pmid"] = pd.NA

            except Exception as e:
                metadata_df["pmid"] = pd.NA

        # Filter out XML namespace artifacts and replace with NA
        metadata_df = metadata_df.replace(
            regex=r"^@xmlns.*", value=pd.NA
        ).infer_objects(copy=False)

        if "run_accession" in metadata_df.columns:
            return metadata_df.sort_values(by="run_accession")
        return metadata_df

    def fetch_gds_results(self, gse, **kwargs):
        result = self.get_esummary_response("geo", gse)

        try:
            uids = result["uids"]
        except KeyError:
            print("No results found for {} | Obtained result: {}".format(gse, result))
            return None
        gse_records = []
        for uid in uids:
            record = result[uid]
            del record["uid"]
            if record["extrelations"]:
                extrelations = record["extrelations"]
                for extrelation in extrelations:
                    keys = list(extrelation.keys())
                    values = list(extrelation.values())
                    assert sorted(keys) == sorted(
                        ["relationtype", "targetobject", "targetftplink"]
                    )
                    assert len(values) == 3
                    record[extrelation["relationtype"]] = extrelation["targetobject"]
                del record["extrelations"]
                gse_records.append(record)
        if not len(gse_records):
            print("No results found for {}".format(gse))
            return None
        return pd.DataFrame(gse_records)

    def gse_to_gsm(self, gse, **kwargs):
        if isinstance(gse, str):
            gse = [gse]
        gse_df = self.fetch_gds_results(gse, **kwargs)
        gse_df = gse_df.rename(
            columns={
                "accession": "experiment_alias",
                "SRA": "experiment_accession",
                "title": "experiment_title",
                "summary": "sample_attribute",
            }
        )
        # TODO: Fix for multiple GSEs?
        gse_df["study_alias"] = ""
        if len(gse) == 1:
            study_alias = gse[0]
        for index, row in gse_df.iterrows():
            if row.entrytype == "GSE":
                study_alias = row["experiment_accession"]
            # If GSM is ecnountered, apply it the
            # previously encountered GSE
            elif row.entrytype == "GSM":
                gse_df.loc[index, "study_alias"] = study_alias
        gse_df = gse_df[gse_df.entrytype == "GSM"]
        if kwargs and kwargs["detailed"] == True:
            return gse_df
        return gse_df[
            ["study_alias", "experiment_alias", "experiment_accession"]
        ].drop_duplicates()

    def gse_to_srp(self, gse, **kwargs):
        if isinstance(gse, str):
            gse = [gse]
        if not gse:  # Handle empty input
            return pd.DataFrame(columns=["study_alias", "study_accession"])

        gse_df = self.fetch_gds_results(gse, **kwargs)
        if gse_df is None or gse_df.empty:  # Handle case where no results found
            return pd.DataFrame(columns=["study_alias", "study_accession"])

        gse_df = gse_df.rename(
            columns={"accession": "study_alias", "SRA": "study_accession"}
        )
        gse_df_subset = None
        if "GSE" in gse_df.entrytype.unique():
            gse_df_subset = gse_df[gse_df.entrytype == "GSE"]
            common_gses = set(gse_df.study_alias.unique()).intersection(gse)
            if len(common_gses) < len(gse):
                gse_df_subset = None
        if gse_df_subset is None:
            # sometimes SRX ids are returned instead of an entire project
            # see https://github.com/saketkc/pysradb/issues/186
            # GSE: GSE209835; SRP =SRP388275
            gse_df_subset_gse = gse_df[gse_df.entrytype == "GSE"]
            gse_of_interest = list(set(gse).difference(gse_df.study_alias.unique()))
            gse_df_subset_other = gse_df[gse_df.entrytype != "GSE"]
            srx = gse_df_subset_other.study_accession.tolist()
            srp_df = self.srx_to_srp(srx)
            srp_unique = list(
                set(srp_df.study_accession.unique()).difference(
                    gse_df_subset_gse.study_accession.tolist()
                )
            )
            # Handle mismatched lengths between GSEs and SRPs
            # Create all combinations of GSE-SRP pairs
            gse_srp_pairs = []
            for gse_id in gse_of_interest:
                for srp_id in srp_unique:
                    gse_srp_pairs.append(
                        {"study_alias": gse_id, "study_accession": srp_id}
                    )

            if gse_srp_pairs:
                new_gse_df = pd.DataFrame(gse_srp_pairs)
            else:
                # If no pairs, create empty DataFrame with correct columns
                new_gse_df = pd.DataFrame(columns=["study_alias", "study_accession"])
            gse_df_subset = pd.concat([gse_df_subset_gse, new_gse_df])
        gse_df_subset = gse_df_subset.loc[gse_df_subset.study_alias.isin(gse)]
        return gse_df_subset[["study_alias", "study_accession"]].drop_duplicates()

    def gsm_to_srp(self, gsm, **kwargs):
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSE"]
        gsm_df = gsm_df.rename(
            columns={"accession": "experiment_alias", "SRA": "study_accession"}
        )
        return gsm_df[["experiment_alias", "study_accession"]].drop_duplicates()

    def gsm_to_srr(self, gsm, **kwargs):
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df.rename(
            columns={
                "accession": "experiment_alias",
                "SRA": "experiment_accession",
                "title": "experiment_title",
                "summary": "sample_attribute",
            }
        )
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"]
        srr_df = self.srx_to_srr(gsm_df.experiment_accession.tolist())
        gsm_df = gsm_df.merge(srr_df, on="experiment_accession")
        return gsm_df[["experiment_alias", "run_accession"]]

    def gsm_to_srs(self, gsm, **kwargs):
        """Get SRS for a GSM"""
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"].rename(
            columns={"SRA": "experiment_accession", "accession": "experiment_alias"}
        )
        srx = gsm_df.experiment_accession.tolist()
        time.sleep(self.sleep_time)
        srs_df = self.srx_to_srs(srx)
        gsm_df = srs_df.merge(gsm_df, on="experiment_accession")[
            ["experiment_alias", "sample_accession"]
        ]
        return gsm_df.drop_duplicates()

    def gsm_to_srx(self, gsm, **kwargs):
        """Get SRX for a GSM"""
        if isinstance(gsm, str):
            gsm = [gsm]
        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"].rename(
            columns={"SRA": "experiment_accession", "accession": "experiment_alias"}
        )
        gsm_df = gsm_df.loc[gsm_df["experiment_alias"].isin(gsm)]
        return gsm_df[["experiment_alias", "experiment_accession"]].drop_duplicates()

    def gsm_to_gse(self, gsm, **kwargs):
        if isinstance(gsm, str):
            gsm = [gsm]
        if not gsm:  # Handle empty input
            return pd.DataFrame(columns=["study_alias", "study_accession"])

        gsm_df = self.fetch_gds_results(gsm, **kwargs)
        if gsm_df is None or gsm_df.empty:  # Handle case where no results found
            return pd.DataFrame(columns=["study_alias", "study_accession"])

        # For GSM queries, we need to extract GSE IDs from the 'gse' column
        # The entrytype will be 'GSM', not 'GSE'
        gsm_entries = gsm_df[gsm_df.entrytype == "GSM"]

        if gsm_entries.empty:
            return pd.DataFrame(columns=["study_alias", "study_accession"])

        # Extract GSE IDs from the 'gse' column and create result rows
        results = []
        for _, row in gsm_entries.iterrows():
            gsm_id = row["accession"]
            gse_str = str(row.get("gse", ""))

            if gse_str and gse_str != "nan":
                # Handle multiple GSE IDs separated by semicolon
                gse_ids = [
                    gse_id.strip() for gse_id in gse_str.split(";") if gse_id.strip()
                ]
                for gse_id in gse_ids:
                    if gse_id.isdigit():
                        # Add GSE prefix if it's just a number
                        gse_id = f"GSE{gse_id}"
                    results.append(
                        {
                            "study_alias": gse_id,
                            "study_accession": row.get("SRA", pd.NA),
                        }
                    )

        if results:
            result_df = pd.DataFrame(results)
            return result_df[["study_alias", "study_accession"]].drop_duplicates()
        else:
            return pd.DataFrame(columns=["study_alias", "study_accession"])

    def srp_to_gse(self, srp, **kwargs):
        """Get GSE for a SRP"""
        srp_df = self.fetch_gds_results(srp, **kwargs)
        if srp_df is None:
            srp_df = pd.DataFrame(
                {"study_alias": [], "study_accession": [], "entrytype": []}
            )

        srp_df = srp_df.rename(
            columns={"accession": "study_alias", "SRA": "study_accession"}
        )
        srp_df_gse = srp_df[srp_df.entrytype == "GSE"]
        missing_srp = list(set(srp).difference(srp_df_gse.study_accession.tolist()))
        srp_df_nongse = srp_df[srp_df.entrytype != "GSE"]
        if srp_df_nongse.shape[0] >= 1:
            srp_df_nongse = pd.DataFrame(
                {
                    "study_accession": missing_srp,
                    "study_alias": [pd.NA] * len(missing_srp),
                    "entrytpe": ["GSE"] * len(missing_srp),
                }
            )
        srp_df = pd.concat([srp_df_gse, srp_df_nongse])
        return srp_df[["study_accession", "study_alias"]].drop_duplicates()

    def srp_to_srr(self, srp, **kwargs):
        """Get SRR for a SRP"""
        srp_df = self.sra_metadata(srp, **kwargs)
        return _order_first(srp_df, ["study_accession", "run_accession"])

    def srp_to_srs(self, srp, **kwargs):
        """Get SRS for a SRP"""
        srp_df = self.sra_metadata(srp, **kwargs)
        return _order_first(srp_df, ["study_accession", "sample_accession"])

    def srp_to_srx(self, srp, **kwargs):
        """Get SRX for a SRP"""
        srp_df = self.sra_metadata(srp, **kwargs)
        srp_df["study_accesssion"] = srp
        return _order_first(srp_df, ["study_accession", "experiment_accession"])

    def srr_to_gsm(self, srr, **kwargs):
        """Get GSM for a SRR"""
        if isinstance(srr, str):
            srr = [srr]
        srr_df = self.srr_to_srp(srr, detailed=True)
        # remove NAs
        srp = [x for x in srr_df.study_accession.tolist() if not x is pd.NA]
        gse_df = self.fetch_gds_results(srp, **kwargs)
        gse_df = gse_df[gse_df.entrytype == "GSE"].rename(
            columns={"SRA": "project_accession", "accession": "project_alias"}
        )
        gsm_df = self.gse_to_gsm(gse_df.project_alias.tolist(), detailed=True)
        srr_cols = list(
            set(srr_df.columns.tolist()).difference(gsm_df.columns.tolist())
        ) + ["experiment_accession"]
        joined_df = gsm_df.merge(srr_df[srr_cols], on="experiment_accession")
        df = _order_first(joined_df, ["run_accession", "experiment_alias"])
        df = df.loc[df["run_accession"].isin(srr)]
        return df

    def srr_to_srp(self, srr, **kwargs):
        """Get SRP for a SRR"""
        if isinstance(srr, str):
            srr = [srr]
        srr_df = self.sra_metadata(srr, **kwargs)
        if kwargs and kwargs["detailed"] == True:
            return srr_df
        srr_df = srr_df.loc[srr_df["run_accession"].isin(srr)]
        return _order_first(srr_df, ["run_accession", "study_accession"])

    def srr_to_srs(self, srr, **kwargs):
        """Get SRS for a SRR"""
        if isinstance(srr, str):
            srr = [srr]
        srr_df = self.sra_metadata(srr, **kwargs)
        srr_df = srr_df.loc[srr_df["run_accession"].isin(srr)]
        return _order_first(srr_df, ["run_accession", "sample_accession"])

    def srr_to_srx(self, srr, **kwargs):
        """Get SRX for a SRR"""
        if isinstance(srr, str):
            srr = [srr]
        srr_df = self.sra_metadata(srr)
        srr_df = srr_df.loc[srr_df["run_accession"].isin(srr)]
        return _order_first(srr_df, ["run_accession", "experiment_accession"])

    def srs_to_gsm(self, srs, **kwargs):
        """Get GSM for a SRS"""
        if isinstance(srs, str):
            srs = [srs]
        srx_df = self.srs_to_srx(srs)
        time.sleep(self.sleep_time)
        gsm_df = self.srx_to_gsm(srx_df.experiment_accession.tolist(), **kwargs)
        srs_df = srx_df.merge(gsm_df, on="experiment_accession")
        srs_df = srs_df.loc[srs_df["sample_accession"].isin(srs)]
        return _order_first(srs_df, ["sample_accession", "experiment_alias"])

    def srx_to_gsm(self, srx, **kwargs):
        if isinstance(srx, str):
            srx = [srx]
        gsm_df = self.fetch_gds_results(srx, **kwargs)
        gsm_df = gsm_df[gsm_df.entrytype == "GSM"].rename(
            columns={"SRA": "experiment_accession", "accession": "experiment_alias"}
        )
        gsm_df = gsm_df.loc[gsm_df["experiment_accession"].isin(srx)]
        return gsm_df[["experiment_accession", "experiment_alias"]].drop_duplicates()

    def srs_to_srx(self, srs, **kwargs):
        """Get SRX for a SRS"""
        srs_df = self.sra_metadata(srs, **kwargs)
        return _order_first(srs_df, ["sample_accession", "experiment_accession"])

    def srx_to_srp(self, srx, **kwargs):
        """Get SRP for a SRX"""
        srx_df = self.sra_metadata(srx, **kwargs)
        return _order_first(srx_df, ["experiment_accession", "study_accession"])

    def srx_to_srr(self, srx, **kwargs):
        """Get SRR for a SRX"""
        srx_df = self.sra_metadata(srx, **kwargs)
        return _order_first(srx_df, ["experiment_accession", "run_accession"])

    def srx_to_srs(self, srx, **kwargs):
        """Get SRS for a SRX"""
        srx_df = self.sra_metadata(srx, **kwargs)
        return _order_first(srx_df, ["experiment_accession", "sample_accession"])

    def search(self, *args, **kwargs):
        raise NotImplementedError("Search not yet implemented for Web")

    def fetch_bioproject_pmids(self, bioprojects):
        """Fetch PMIDs for given BioProject accessions

        Parameters
        ----------
        bioprojects: list or str
                    BioProject accession(s)

        Returns
        -------
        bioproject_pmids: dict
                         Mapping of BioProject to list of PMIDs
        """
        if isinstance(bioprojects, str):
            bioprojects = [bioprojects]

        bioproject_pmids = {}
        for bioproject in bioprojects:
            if pd.isna(bioproject) or not bioproject:
                bioproject_pmids[bioproject] = []
                continue

            try:
                payload = self.efetch_params.copy()
                payload = [param for param in payload if param[0] != "retmode"]
                payload += [
                    ("db", "bioproject"),
                    ("id", bioproject),
                    ("retmode", "xml"),
                ]

                request = requests.get(
                    self.base_url["efetch"], params=OrderedDict(payload)
                )
                xml_text = request.text.strip()

                # Parse XML to extract Publication IDs
                pmids = []
                try:
                    xml_dict = xmltodict.parse(
                        xml_text, process_namespaces=False, dict_constructor=OrderedDict
                    )

                    # Navigate through the XML structure
                    if "RecordSet" in xml_dict:
                        records = xml_dict["RecordSet"].get("DocumentSummary", [])
                        if not isinstance(records, list):
                            records = [records]

                        for record in records:
                            project_descr = record.get("Project", {}).get(
                                "ProjectDescr", {}
                            )
                            publications = project_descr.get("Publication", [])

                            if not isinstance(publications, list):
                                publications = [publications]

                            for pub in publications:
                                pub_id = pub.get("@id", "")
                                if pub_id and pub_id.isdigit():
                                    pmids.append(pub_id)

                except ExpatError:
                    # XML parsing failed --> Look for PMID patterns in the raw text
                    pmid_pattern = r'id="(\d+)"'
                    matches = re.findall(pmid_pattern, xml_text)
                    pmids = [
                        match for match in matches if len(match) >= 7
                    ]  # PMIDs are typically 7+ digits

                bioproject_pmids[bioproject] = list(set(pmids))  # Remove duplicates
                time.sleep(self.sleep_time)

            except Exception as e:
                warnings.warn(f"Failed to fetch PMIDs for BioProject {bioproject}: {e}")
                bioproject_pmids[bioproject] = []

        return bioproject_pmids

    def srp_to_pmid(self, srp_accessions):
        """Get PMIDs associated with SRP accessions

        Parameters
        ----------
        srp_accessions: list or str
                       SRP accession(s)

        Returns
        -------
        srp_pmid_df: pandas.DataFrame
                    DataFrame with SRP accessions and associated PMIDs
        """
        if isinstance(srp_accessions, str):
            srp_accessions = [srp_accessions]

        # Get metadata to extract BioProject information
        metadata_df = self.sra_metadata(srp_accessions)
        if metadata_df is None or metadata_df.empty:
            return pd.DataFrame(columns=["srp_accession", "bioproject", "pmid"])

        # Try to get PMIDs via BioProject first
        unique_bioprojects = metadata_df["bioproject"].dropna().unique().tolist()
        bioproject_pmids = self.fetch_bioproject_pmids(unique_bioprojects)

        # If no BioProject PMIDs found, try fallback search
        external_pmids = []
        if not any(pmids for pmids in bioproject_pmids.values()):
            external_pmids = self._search_fallback_pmids(srp_accessions)

        # Build results - one row per unique SRP accession
        results = []
        for _, row in metadata_df.iterrows():
            srp_acc = self._extract_sra_accession(row)
            bioproject = row.get("bioproject", "")

            # Get PMIDs (BioProject takes priority over external)
            pmids = bioproject_pmids.get(bioproject, [])
            if not pmids and external_pmids:
                pmids = external_pmids

            # Add result with smallest PMID (if any found)
            smallest_pmid = self._get_smallest_pmid(pmids) if pmids else pd.NA
            results.append(
                {
                    "srp_accession": srp_acc,
                    "bioproject": bioproject,
                    "pmid": smallest_pmid,
                }
            )

        return pd.DataFrame(results).drop_duplicates()

    def _search_fallback_pmids(self, srp_accessions):
        """Search for PMIDs using fallback strategies (external sources + direct SRA search + GSE search)"""
        try:
            original_sleep = self.sleep_time
            self.sleep_time = max(0.1, self.sleep_time * 0.5)

            # Strategy 1: Search via external source identifiers
            # Example: ERP018009
            detailed_metadata = self.sra_metadata(
                srp_accessions, detailed=True, include_pmids=False
            )
            if detailed_metadata is not None and not detailed_metadata.empty:
                if external_sources := self.extract_external_sources(detailed_metadata):
                    pmids = self.search_pmc_for_external_sources([external_sources[0]])
                    if pmids:
                        return pmids

                # Strategy 2: Search via GSE identifiers extracted from metadata
                # Example: GSE253406 --> SRP484103
                gse_pmids = self._search_gse_gsm_pmids(
                    detailed_metadata, srp_accessions
                )
                if gse_pmids:
                    return gse_pmids

            # Strategy 3: Direct SRP ID search
            # Example: SRP047086
            pmids = self.search_pmc_for_external_sources(srp_accessions)
            return pmids

        except Exception as e:
            return []
        finally:
            self.sleep_time = original_sleep

    def _extract_sra_accession(self, row):
        """Extract SRA accession from metadata row"""
        return row.get(
            "study_accession",
            row.get(
                "run_accession",
                row.get("experiment_accession", row.get("sample_accession", "")),
            ),
        )

    def _get_smallest_pmid(self, pmids):
        """Get the numerically smallest PMID from a list"""
        if not pmids:
            return pd.NA

        # Convert to integers for proper numeric sorting
        pmid_ints = []
        for pmid in pmids:
            try:
                pmid_ints.append(int(pmid))
            except ValueError:
                pmid_ints.append(pmid)  # Keep non-numeric as-is

        return str(min(pmid_ints))

    def extract_external_sources(self, metadata_df):
        """Extract external source identifiers from SRA metadata

        Parameters
        ----------
        metadata_df: pandas.DataFrame
                    DataFrame containing SRA metadata

        Returns
        -------
        external_sources: list
                         List of external source identifiers found
        """
        external_sources = []

        patterns = [
            r"E-MTAB-\d+",  # ArrayExpress
            r"GSE\d+",  # GEO Series
            r"E-GEOD-\d+",  # GEO in ArrayExpress
            r"E-MEXP-\d+",  # MEXP in ArrayExpress
            r"E-TABM-\d+",  # TABM in ArrayExpress
        ]

        # Fields that commonly contain external source identifiers
        source_fields = ["run_alias", "submitter id", "sample name", "experiment_alias"]

        for field in source_fields:
            if field in metadata_df.columns:
                values = metadata_df[field].dropna().unique()
                for value in values:
                    value_str = str(value)
                    for pattern in patterns:
                        matches = re.findall(pattern, value_str)
                        external_sources.extend(
                            match for match in matches if match not in external_sources
                        )

        return external_sources

    def _search_gse_gsm_pmids(self, metadata_df, sra_accessions):
        """Search for PMIDs using GSE identifiers from BioProject and SRP conversion

        Parameters
        ----------
        metadata_df: pandas.DataFrame
                    Detailed metadata DataFrame
        sra_accessions: list
                       List of SRA accessions being searched

        Returns
        -------
        pmids: list
              List of PMIDs found via GSE search
        """
        import time

        gse_identifiers = []

        # Strategy 1: BioProject to GSE conversion via NCBI search
        if "bioproject" in metadata_df.columns:
            unique_bioprojects = metadata_df["bioproject"].dropna().unique()
            for bioproject in unique_bioprojects[
                :3
            ]:  # Limit to avoid too many requests
                try:
                    gse_ids = self._bioproject_to_gse(bioproject)
                    gse_identifiers.extend(gse_ids)
                    time.sleep(self.sleep_time)  # Rate limiting
                except Exception:
                    pass

        # Strategy 2: SRP to GSE conversion via NCBI ELink
        for sra_acc in sra_accessions:
            if sra_acc.startswith("SRP"):
                try:
                    gse_ids = self._srp_to_gse_via_elink(sra_acc)
                    gse_identifiers.extend(gse_ids)
                    time.sleep(self.sleep_time)  # Rate limiting
                except Exception:
                    pass

        # Strategy 3: Try existing pysradb SRP to GSE conversion
        for sra_acc in sra_accessions:
            if sra_acc.startswith("SRP"):
                try:
                    gse_df = self.srp_to_gse(sra_acc)
                    if not gse_df.empty and "experiment_alias" in gse_df.columns:
                        gse_values = gse_df["experiment_alias"].dropna().astype(str)
                        for gse_val in gse_values:
                            if gse_val.startswith("GSE"):
                                gse_identifiers.append(gse_val)
                except Exception:
                    pass

        # Remove duplicates and search PMC for GSE identifiers
        unique_gse_ids = list(set(gse_identifiers))

        if unique_gse_ids:
            pmids = self.search_pmc_for_external_sources(unique_gse_ids)
            return pmids

        return []

    def _bioproject_to_gse(self, bioproject):
        """Convert BioProject ID to GSE ID via NCBI search

        Parameters
        ----------
        bioproject: str
                   BioProject ID (e.g., 'PRJNA1065472')

        Returns
        -------
        gse_ids: list
                List of GSE IDs found
        """
        import requests

        gse_ids = []
        try:
            search_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
            search_params = {
                "db": "gds",
                "term": f"{bioproject}[BioProject]",
                "retmode": "json",
                "retmax": "10",
            }

            response = requests.get(search_url, params=search_params, timeout=30)
            response.raise_for_status()
            result = response.json()
            geo_uids = result["esearchresult"]["idlist"]

            if geo_uids:
                # Get summary to find GSE IDs
                summary_url = (
                    "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi"
                )
                summary_params = {
                    "db": "gds",
                    "id": ",".join(geo_uids),
                    "retmode": "json",
                }

                summary_response = requests.get(
                    summary_url, params=summary_params, timeout=30
                )
                summary_response.raise_for_status()
                summary_result = summary_response.json()

                for uid in geo_uids:
                    if uid in summary_result["result"]:
                        record = summary_result["result"][uid]
                        accession = record.get("accession", "")
                        if accession.startswith("GSE"):
                            gse_ids.append(accession)

        except Exception:
            pass

        return gse_ids

    def _srp_to_gse_via_elink(self, srp_id):
        """Convert SRP ID to GSE ID via NCBI ELink

        Parameters
        ----------
        srp_id: str
               SRP ID (e.g., 'SRP484103')

        Returns
        -------
        gse_ids: list
                List of GSE IDs found
        """
        import requests

        gse_ids = []
        try:
            # First, search for the SRP in SRA database to get UIDs
            search_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
            search_params = {
                "db": "sra",
                "term": srp_id,
                "retmode": "json",
                "retmax": "5",
            }

            response = requests.get(search_url, params=search_params, timeout=30)
            response.raise_for_status()
            result = response.json()
            sra_uids = result["esearchresult"]["idlist"]

            if sra_uids:
                # Use ELink to find related GEO records
                elink_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi"
                elink_params = {
                    "dbfrom": "sra",
                    "db": "gds",
                    "id": sra_uids[0],  # Use first UID
                    "retmode": "json",
                }

                elink_response = requests.get(
                    elink_url, params=elink_params, timeout=30
                )
                elink_response.raise_for_status()
                elink_result = elink_response.json()

                if "linksets" in elink_result:
                    for linkset in elink_result["linksets"]:
                        if "linksetdbs" in linkset:
                            for linksetdb in linkset["linksetdbs"]:
                                if linksetdb["dbto"] == "gds":
                                    geo_uids = linksetdb["links"]

                                    if geo_uids:
                                        # Get summary to find GSE IDs
                                        summary_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi"
                                        summary_params = {
                                            "db": "gds",
                                            "id": ",".join(geo_uids),
                                            "retmode": "json",
                                        }

                                        summary_response = requests.get(
                                            summary_url,
                                            params=summary_params,
                                            timeout=30,
                                        )
                                        summary_response.raise_for_status()
                                        summary_result = summary_response.json()

                                        for uid in geo_uids:
                                            if uid in summary_result["result"]:
                                                record = summary_result["result"][uid]
                                                accession = record.get("accession", "")
                                                if accession.startswith("GSE"):
                                                    gse_ids.append(accession)

        except Exception:
            pass

        return gse_ids

    def search_pmc_for_external_sources(self, external_sources):
        """Search PubMed Central for PMIDs using external source identifiers

        Parameters
        ----------
        external_sources: list
                         List of external source identifiers

        Returns
        -------
        pmids: list
              List of PMIDs found
        """
        if not external_sources:
            return []

        all_pmids = []

        for source in external_sources:
            try:
                search_url = (
                    "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
                )
                search_params = {
                    "db": "pmc",
                    "term": source,
                    "retmode": "json",
                    "retmax": "10",
                }

                response = requests.get(search_url, params=search_params, timeout=60)
                response.raise_for_status()
                result = response.json()

                pmc_ids = result["esearchresult"]["idlist"]
                if not pmc_ids:
                    continue

                # Get primary PMIDs for each PMC article
                summary_url = (
                    "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi"
                )
                summary_params = {
                    "db": "pmc",
                    "id": ",".join(pmc_ids),
                    "retmode": "json",
                }

                summary_response = requests.get(
                    summary_url, params=summary_params, timeout=60
                )
                summary_result = summary_response.json()

                # Extract primary PMID for each PMC article
                for pmc_id in pmc_ids:
                    if pmc_id in summary_result["result"]:
                        article = summary_result["result"][pmc_id]
                        articleids = article.get("articleids", [])

                        # Find the primary PMID
                        for aid in articleids:
                            if aid.get("idtype") == "pmid":
                                primary_pmid = aid.get("value")
                                if primary_pmid and primary_pmid not in all_pmids:
                                    all_pmids.append(primary_pmid)
                                break

                time.sleep(self.sleep_time)  # Rate limiting

            except Exception as e:
                continue

        return list(set(all_pmids))  # Remove duplicates

    def sra_to_pmid(self, sra_accessions):
        """Get PMIDs for SRA accessions (backward compatibility wrapper)

        Parameters
        ----------
        sra_accessions: list or str
                       SRA accession(s) - can be SRP, SRR, SRX, or SRS

        Returns
        -------
        sra_pmid_df: pandas.DataFrame
                    DataFrame with SRA accessions and associated PMIDs
        """
        # For SRP accessions, use the main method
        if isinstance(sra_accessions, str):
            if sra_accessions.startswith("SRP"):
                return self.srp_to_pmid(sra_accessions)
        elif isinstance(sra_accessions, list):
            # If all are SRP accessions, use main method
            if all(acc.startswith("SRP") for acc in sra_accessions):
                return self.srp_to_pmid(sra_accessions)

        # For other SRA accessions, convert to SRP first if possible
        # This is a simplified implementation for backward compatibility
        return self.srp_to_pmid(sra_accessions)

    def srr_to_pmid(self, srr):
        """Get PMIDs for Run Accessions (SRR)"""
        return self.sra_to_pmid(srr)

    def srx_to_pmid(self, srx):
        """Get PMIDs for Experiment Accessions (SRX)"""
        return self.sra_to_pmid(srx)

    def srs_to_pmid(self, srs):
        """Get PMIDs for Sample Accessions (SRS)"""
        return self.sra_to_pmid(srs)

    def gse_to_pmid(self, gse_accessions):
        """Get PMIDs for GSE accessions by searching PubMed Central

        Parameters
        ----------
        gse_accessions: list or str
                       GSE accession(s)

        Returns
        -------
        gse_pmid_df: pandas.DataFrame
                    DataFrame with GSE accessions and associated PMIDs
        """
        if isinstance(gse_accessions, str):
            gse_accessions = [gse_accessions]

        results = []
        for gse_acc in gse_accessions:
            pmids = self.search_pmc_for_external_sources([gse_acc])
            smallest_pmid = self._get_smallest_pmid(pmids) if pmids else pd.NA

            results.append(
                {
                    "gse_accession": gse_acc,
                    "pmid": smallest_pmid,
                }
            )

        return pd.DataFrame(results)

    def doi_to_pmid(self, dois):
        """Convert DOI(s) to PMID(s)

        Parameters
        ----------
        dois: list or str
             DOI(s)

        Returns
        -------
        doi_pmid_mapping: dict
                         Mapping of DOI to PMID
        """
        if isinstance(dois, str):
            dois = [dois]

        doi_pmid_mapping = {}

        for doi in dois:
            try:
                search_url = self.base_url["esearch"]
                search_params = {
                    "db": "pubmed",
                    "term": f"{doi}[DOI]",
                    "retmode": "json",
                }

                response = requests.get(search_url, params=search_params, timeout=60)
                response.raise_for_status()
                result = response.json()

                id_list = result.get("esearchresult", {}).get("idlist", [])
                if id_list:
                    doi_pmid_mapping[doi] = id_list[0]
                else:
                    doi_pmid_mapping[doi] = None

                time.sleep(self.sleep_time)

            except requests.RequestException as e:
                warnings.warn(f"Network error while getting PMID for DOI {doi}: {e}")
                doi_pmid_mapping[doi] = None
            except ValueError as e:
                warnings.warn(
                    f"Value error while processing response for DOI {doi}: {e}"
                )
                doi_pmid_mapping[doi] = None

        return doi_pmid_mapping

    def pmid_to_pmc(self, pmids):
        """Convert PMID(s) to PMC ID(s)

        Parameters
        ----------
        pmids: list or str
              PMID(s)

        Returns
        -------
        pmid_pmc_mapping: dict
                         Mapping of PMID to PMC ID
        """
        if isinstance(pmids, str):
            pmids = [pmids]

        pmid_pmc_mapping = {}

        for pmid in pmids:
            try:
                summary_url = self.base_url["esummary"]
                summary_params = {
                    "db": "pubmed",
                    "id": pmid,
                    "retmode": "json",
                }

                response = requests.get(summary_url, params=summary_params, timeout=60)
                response.raise_for_status()
                result = response.json()

                # Extract PMC ID from articleids
                if str(pmid) in result.get("result", {}):
                    article = result["result"][str(pmid)]
                    articleids = article.get("articleids", [])

                    for aid in articleids:
                        if aid.get("idtype") == "pmc":
                            pmc_id = aid.get("value")
                            pmid_pmc_mapping[pmid] = pmc_id
                            break

                time.sleep(self.sleep_time)

            except Exception as e:
                warnings.warn(f"Failed to get PMC ID for PMID {pmid}: {e}")
                pmid_pmc_mapping[pmid] = None

        return pmid_pmc_mapping

    def fetch_pmc_fulltext(self, pmc_id):
        """Fetch full text from PMC article

        Parameters
        ----------
        pmc_id: str
               PMC ID (can be with or without 'PMC' prefix)

        Returns
        -------
        fulltext: str
                 Full text of the article, or None if unavailable
        """
        # Ensure PMC ID has the PMC prefix
        if not pmc_id.startswith("PMC"):
            pmc_id = f"PMC{pmc_id}"

        try:
            fetch_url = self.base_url["efetch"]
            fetch_params = {"db": "pmc", "id": pmc_id, "retmode": "xml"}

            response = requests.get(fetch_url, params=fetch_params, timeout=60)
            response.raise_for_status()

            time.sleep(self.sleep_time)
            return response.text

        except Exception as e:
            warnings.warn(f"Failed to fetch full text for {pmc_id}: {e}")
            return None

    def extract_identifiers_from_text(self, text):
        """Extract GSE, PRJNA, SRP, and other identifiers from text

        Parameters
        ----------
        text: str
             Text to search for identifiers

        Returns
        -------
        identifiers: dict
                    Dictionary with lists of found identifiers by type
        """
        if not text:
            return {
                "gse": [],
                "prjna": [],
                "srp": [],
                "srr": [],
                "srx": [],
                "srs": [],
            }

        # Define patterns for different identifier types
        patterns = {
            "gse": r"GSE\d+",
            "prjna": r"PRJNA\d+",
            "srp": r"SRP\d+",
            "srr": r"SRR\d+",
            "srx": r"SRX\d+",
            "srs": r"SRS\d+",
        }

        identifiers = {}
        for id_type, pattern in patterns.items():
            matches = re.findall(pattern, text, re.IGNORECASE)
            # Convert to uppercase and remove duplicates
            identifiers[id_type] = sorted(list(set([m.upper() for m in matches])))

        return identifiers

    def pmc_to_identifiers(self, pmc_ids, convert_missing=True):
        """Extract database identifiers from PMC articles

        Parameters
        ----------
        pmc_ids: list or str
                PMC ID(s) (can be with or without 'PMC' prefix)
        convert_missing: bool
                        If True, automatically convert GSE↔SRP when one is found but not the other
                        Default: True

        Returns
        -------
        results_df: pandas.DataFrame
                   DataFrame with PMC IDs and extracted identifiers
        """
        if isinstance(pmc_ids, str):
            pmc_ids = [pmc_ids]

        results = []

        for pmc_id in pmc_ids:
            # Fetch full text
            fulltext = self.fetch_pmc_fulltext(pmc_id)

            if fulltext:
                # Extract identifiers
                identifiers = self.extract_identifiers_from_text(fulltext)

                if convert_missing:
                    # If we found GSE IDs but no SRP IDs, convert GSE to SRP
                    if identifiers["gse"] and not identifiers["srp"]:
                        try:
                            for gse_id in identifiers["gse"]:
                                gse_srp_df = self.gse_to_srp(gse_id)
                                if (
                                    not gse_srp_df.empty
                                    and "study_accession" in gse_srp_df.columns
                                ):
                                    srp_values = (
                                        gse_srp_df["study_accession"].dropna().tolist()
                                    )
                                    identifiers["srp"].extend(
                                        [str(x) for x in srp_values if not pd.isna(x)]
                                    )
                            identifiers["srp"] = sorted(list(set(identifiers["srp"])))
                            time.sleep(self.sleep_time)
                        except Exception:
                            pass

                    # If we found SRP IDs but no GSE IDs, convert SRP to GSE
                    elif identifiers["srp"] and not identifiers["gse"]:
                        try:
                            for srp_id in identifiers["srp"]:
                                srp_gse_df = self.srp_to_gse(srp_id)
                                if (
                                    not srp_gse_df.empty
                                    and "study_alias" in srp_gse_df.columns
                                ):
                                    gse_values = (
                                        srp_gse_df["study_alias"].dropna().tolist()
                                    )
                                    identifiers["gse"].extend(
                                        [str(x) for x in gse_values if not pd.isna(x)]
                                    )
                            identifiers["gse"] = sorted(
                                list(set(identifiers["gse"]))
                            )  # Remove duplicates
                            time.sleep(self.sleep_time)
                        except Exception:
                            pass  # If conversion fails, just keep what we found

                results.append(
                    {
                        "pmc_id": (
                            pmc_id if pmc_id.startswith("PMC") else f"PMC{pmc_id}"
                        ),
                        "gse_ids": (
                            ",".join(identifiers["gse"])
                            if identifiers["gse"]
                            else pd.NA
                        ),
                        "prjna_ids": (
                            ",".join(identifiers["prjna"])
                            if identifiers["prjna"]
                            else pd.NA
                        ),
                        "srp_ids": (
                            ",".join(identifiers["srp"])
                            if identifiers["srp"]
                            else pd.NA
                        ),
                        "srr_ids": (
                            ",".join(identifiers["srr"])
                            if identifiers["srr"]
                            else pd.NA
                        ),
                        "srx_ids": (
                            ",".join(identifiers["srx"])
                            if identifiers["srx"]
                            else pd.NA
                        ),
                        "srs_ids": (
                            ",".join(identifiers["srs"])
                            if identifiers["srs"]
                            else pd.NA
                        ),
                    }
                )
            else:
                results.append(
                    {
                        "pmc_id": (
                            pmc_id if pmc_id.startswith("PMC") else f"PMC{pmc_id}"
                        ),
                        "gse_ids": pd.NA,
                        "prjna_ids": pd.NA,
                        "srp_ids": pd.NA,
                        "srr_ids": pd.NA,
                        "srx_ids": pd.NA,
                        "srs_ids": pd.NA,
                    }
                )

        return pd.DataFrame(results)

    def pmid_to_identifiers(self, pmids):
        """Extract database identifiers from PubMed articles via PMC

        Parameters
        ----------
        pmids: list or str
              PMID(s)

        Returns
        -------
        results_df: pandas.DataFrame
                   DataFrame with PMIDs, PMC IDs, and extracted identifiers
        """
        if isinstance(pmids, str):
            pmids = [pmids]

        # First convert PMIDs to PMC IDs
        pmid_pmc_mapping = self.pmid_to_pmc(pmids)

        results = []

        for pmid, pmc_id in pmid_pmc_mapping.items():
            if pmc_id:
                # Get identifiers from PMC
                pmc_results = self.pmc_to_identifiers([pmc_id])

                if not pmc_results.empty:
                    result = pmc_results.iloc[0].to_dict()
                    result["pmid"] = pmid
                    # Reorder columns to have pmid first
                    result = {
                        "pmid": result["pmid"],
                        "pmc_id": result["pmc_id"],
                        "gse_ids": result["gse_ids"],
                        "prjna_ids": result["prjna_ids"],
                        "srp_ids": result["srp_ids"],
                        "srr_ids": result["srr_ids"],
                        "srx_ids": result["srx_ids"],
                        "srs_ids": result["srs_ids"],
                    }
                    results.append(result)
                else:
                    results.append(
                        {
                            "pmid": pmid,
                            "pmc_id": pmc_id,
                            "gse_ids": pd.NA,
                            "prjna_ids": pd.NA,
                            "srp_ids": pd.NA,
                            "srr_ids": pd.NA,
                            "srx_ids": pd.NA,
                            "srs_ids": pd.NA,
                        }
                    )
            else:
                # No PMC ID available
                results.append(
                    {
                        "pmid": pmid,
                        "pmc_id": pd.NA,
                        "gse_ids": pd.NA,
                        "prjna_ids": pd.NA,
                        "srp_ids": pd.NA,
                        "srr_ids": pd.NA,
                        "srx_ids": pd.NA,
                        "srs_ids": pd.NA,
                    }
                )

        return pd.DataFrame(results)

    def pmid_to_gse(self, pmids):
        """Get GSE identifiers from PMID(s)

        Parameters
        ----------
        pmids: list or str
              PMID(s)

        Returns
        -------
        results_df: pandas.DataFrame
                   DataFrame with PMIDs and GSE identifiers
        """
        full_results = self.pmid_to_identifiers(pmids)
        return full_results[["pmid", "pmc_id", "gse_ids"]]

    def pmid_to_srp(self, pmids):
        """Get SRP identifiers from PMID(s)

        Parameters
        ----------
        pmids: list or str
              PMID(s)

        Returns
        -------
        results_df: pandas.DataFrame
                   DataFrame with PMIDs and SRP identifiers
        """
        full_results = self.pmid_to_identifiers(pmids)
        return full_results[["pmid", "pmc_id", "srp_ids"]]

    def doi_to_identifiers(self, dois):
        """Extract database identifiers from articles via DOI

        Parameters
        ----------
        dois: list or str
             DOI(s)

        Returns
        -------
        results_df: pandas.DataFrame
                   DataFrame with DOIs, PMIDs, PMC IDs, and extracted identifiers
        """
        if isinstance(dois, str):
            dois = [dois]

        doi_pmid_mapping = self.doi_to_pmid(dois)

        results = []

        for doi, pmid in doi_pmid_mapping.items():
            if pmid:
                pmid_results = self.pmid_to_identifiers([pmid])

                if not pmid_results.empty:
                    result = pmid_results.iloc[0].to_dict()
                    result["doi"] = doi
                    result = {
                        "doi": result["doi"],
                        "pmid": result["pmid"],
                        "pmc_id": result["pmc_id"],
                        "gse_ids": result["gse_ids"],
                        "prjna_ids": result["prjna_ids"],
                        "srp_ids": result["srp_ids"],
                        "srr_ids": result["srr_ids"],
                        "srx_ids": result["srx_ids"],
                        "srs_ids": result["srs_ids"],
                    }
                    results.append(result)
                else:
                    results.append(
                        {
                            "doi": doi,
                            "pmid": pmid,
                            "pmc_id": pd.NA,
                            "gse_ids": pd.NA,
                            "prjna_ids": pd.NA,
                            "srp_ids": pd.NA,
                            "srr_ids": pd.NA,
                            "srx_ids": pd.NA,
                            "srs_ids": pd.NA,
                        }
                    )
            else:
                # No PMID available
                results.append(
                    {
                        "doi": doi,
                        "pmid": pd.NA,
                        "pmc_id": pd.NA,
                        "gse_ids": pd.NA,
                        "prjna_ids": pd.NA,
                        "srp_ids": pd.NA,
                        "srr_ids": pd.NA,
                        "srx_ids": pd.NA,
                        "srs_ids": pd.NA,
                    }
                )

        return pd.DataFrame(results)

    def doi_to_gse(self, dois):
        """Get GSE identifiers from DOI(s)

        Parameters
        ----------
        dois: list or str
             DOI(s)

        Returns
        -------
        results_df: pandas.DataFrame
                   DataFrame with DOIs and GSE identifiers
        """
        full_results = self.doi_to_identifiers(dois)
        return full_results[["doi", "pmid", "pmc_id", "gse_ids"]]

    def doi_to_srp(self, dois):
        """Get SRP identifiers from DOI(s)

        Parameters
        ----------
        dois: list or str
             DOI(s)

        Returns
        -------
        results_df: pandas.DataFrame
                   DataFrame with DOIs and SRP identifiers
        """
        full_results = self.doi_to_identifiers(dois)
        return full_results[["doi", "pmid", "pmc_id", "srp_ids"]]

    def close(self):
        # Dummy method to mimick SRAdb() object
        pass



"""Methods to interact with SRA"""

# This is now defunct and will be removed in a future release.

import gzip
import os
import re
import sys
from io import StringIO

import pandas as pd

from .basedb import BASEdb
from .utils import _get_url, copyfileobj, get_gzip_uncompressed_size

PY3 = True
if sys.version_info[0] < 3:
    PY3 = False

GEOmetadb_URL = "http://starbuck1.s3.amazonaws.com/sradb/GEOmetadb.sqlite.gz"


def download_geodb_file(download_dir=os.getcwd(), overwrite=True):
    """Download GEOmetadb.sqlite file.

    Parameters
    ----------
    download_dir: string
                  Directory to download SRAmetadb.sqlite
    overwrite: bool
               overwrite existing file(s).
               Set to True by default.

    """
    download_location = os.path.join(download_dir, "GEOmetadb.sqlite.gz")
    download_location_unzip = download_location.rstrip(".gz")

    if os.path.isfile(download_location) and overwrite is False:
        raise RuntimeError(
            "{} already exists! Set `overwrite=True` to redownload.".format(
                download_location
            )
        )
    if os.path.isfile(download_location_unzip) and overwrite is False:
        raise RuntimeError(
            "{} already exists! Set `overwrite=True` to redownload.".format(
                download_location_unzip
            )
        )

    try:
        _get_url(GEOmetadb_URL, download_location)
    except Exception as e:
        raise RuntimeError(
            "Could not use {}.\nException: {}.\n".format(GEOmetadb_URL, e)
        )
    print("Extracting {} ...".format(download_location))
    filesize = get_gzip_uncompressed_size(download_location)
    with gzip.open(download_location, "rb") as fh_in:
        with open(download_location_unzip, "wb") as fh_out:
            copyfileobj(
                fh_in,
                fh_out,
                filesize=filesize,
                desc="Extracting {}".format("GEOmetadb.sqlite.gz"),
            )
    print("Done!")
    db = GEOdb(download_location_unzip)
    metadata = db.query("SELECT * FROM metaInfo")
    db.close()
    print("Metadata associated with {}:".format(download_location_unzip))
    print(metadata)


class GEOdb(BASEdb):
    def __init__(self, sqlite_file):
        """Initialize SRAdb.

        Parameters
        ----------

        sqlite_file: string
                     Path to unzipped SRAmetadb.sqlite file


        """
        super(GEOdb, self).__init__(sqlite_file)
        self._db_type = "GEO"
        self.valid_in_type = ["GSE", "GPL", "GSM", "GDS"]

    def gse_metadata(self, gse):
        """Get metadata for GSE ID.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with relevant fields
        """
        return self.query("SELECT * from gse WHERE gse='{}';".format(gse))

    def gsm_metadata(self, gsm):
        """Get metadata for GSM ID.

        Parameters
        ----------
        gsm: string
             GSM ID

        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with relevant fields
        """
        return self.query("SELECT * from gsm WHERE gsm='{}';".format(gsm))

    def geo_convert(self, from_acc):
        """Convert one GEO accession to other.

        Parameters
        ----------
        from_acc: string
                  GPL/GSE/GSM accession ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        return self.query(
            "SELECT * FROM geoConvert WHERE from_acc='{}';".format(from_acc)
        )

    def gse_to_gsm(self, gse):
        """Fetch GSMs for a GSE.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        return self.query("SELECT * FROM gse_gsm WHERE gse='{}'".format(gse))

    def gsm_to_gse(self, gsm):
        """Fetch GSE for a GSM.

        Parameters
        ----------
        gsm: string
             GSM ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        mapping_df = self.query("SELECT * FROM gse_gsm WHERE gsm='{}'".format(gsm))
        return mapping_df.loc[:, ["gsm", "gse"]]

    def guess_srp_from_gse(self, gse):
        """Convert GSE to SRP id.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        srp: string
             SRP ID
        """
        results = self.query('SELECT * FROM gse WHERE gse = "' + gse + '"')
        if results.shape[0] == 1:
            supp_file = results["supplementary_file"][0]
            if supp_file:
                splitted = supp_file.split(";")
                if len(splitted):
                    match = re.findall("SRP.*", splitted[-1])
                    if len(match):
                        srp = match[0].split("/")[-1]
                        return srp
        return None


def download_geo_matrix(accession, output_dir="."):
    """
    Download a GEO Matrix file for a given GEO accession ID.

    Args:
        accession (str): GEO accession ID (e.g., 'GSE234190').
        output_dir (str): Directory to save the downloaded file (default: current directory).

    Returns:
        str: Path to the downloaded file.

    Raises:
        Exception: If the download fails.
    """
    # Construct the URL for the GEO Matrix file
    url = f"https://ftp.ncbi.nlm.nih.gov/geo/series/{accession[:-3]}nnn/{accession}/matrix/{accession}_series_matrix.txt.gz"

    # Define the output file path
    output_file = os.path.join(output_dir, f"{accession}_series_matrix.txt.gz")

    # Download the file using _get_url
    try:
        _get_url(url, output_file)
        return output_file
    except Exception as e:
        raise Exception(
            f"Failed to download GEO Matrix file for {accession}. Exception: {str(e)}"
        )


def parse_geo_matrix_to_tsv(input_file, output_file):
    """
    Parse a GEO Matrix file to a TSV file, extracting the expression data.

    Args:
        input_file (str): Path to the input GEO Matrix file (gzipped).
        output_file (str): Path to save the output TSV file.

    Returns:
        pandas.DataFrame: The parsed expression data.
    """
    # Read the gzipped file and extract the data section
    data_lines = []
    data_section = False

    with gzip.open(input_file, "rt", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if line == "!series_matrix_table_begin":
                data_section = True
                continue
            elif line == "!series_matrix_table_end":
                break
            if data_section and line:
                data_lines.append(line)

    # Use pandas.read_csv to parse the data section
    df = pd.read_csv(StringIO("\n".join(data_lines)), sep="\t", comment="!")

    # Save to TSV
    df.to_csv(output_file, sep="\t", index=False)

    return df
